# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
The Multi-disciplinary Design Feasible formulation
**************************************************
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library

from gemseo.core.formulation import MDOFormulation
from gemseo.mda.mda_factory import MDAFactory

standard_library.install_aliases()


class MDF(MDOFormulation):
    """
    The Multidisciplinary Design Feasible formulation draws an
    optimization architecture where the coupling of strongly
    coupled disciplines is made consistent by means of a
    Multidisciplinary Design Analysis (MDA), the optimization
    problem w.r.t. local and global design variables is made
    at the top level. Multidisciplinary analysis is made at
    a each optimization iteration.
    """

    def __init__(
        self,
        disciplines,
        objective_name,
        design_space,
        maximize_objective=False,
        main_mda_class="MDAChain",
        sub_mda_class="MDAJacobi",
        **mda_options
    ):
        """
        Constructor, initializes the objective functions and constraints

        :param main_mda_class: classname of the main MDA, typically the
            MDAChain,  but one can force to use MDAGaussSeidel for instance
        :type main_mda_class: str
        :param disciplines: the disciplines list.
        :type disciplines: list(MDODiscipline)
        :param objective_name: the objective function data name.
        :type objective_name: str
        :param design_space: the design space.
        :type design_space: DesignSpace
        :param maximize_objective: if True, the objective function
            is maximized, by default, a minimization is performed.
        :type maximize_objective: bool
        :param sub_mda_class: the type of MDA  to be used,
            shall be the class name. (default MDAJacobi)
        :type sub_mda_class: str
        :param mda_options: options passed to the MDA at construction
        """
        super(MDF, self).__init__(
            disciplines,
            objective_name,
            design_space,
            maximize_objective=maximize_objective,
        )
        self.mda = None
        self._main_mda_class = main_mda_class
        self._mda_factory = MDAFactory()
        self._instantiate_mda(main_mda_class, sub_mda_class, **mda_options)
        self._update_design_space()
        self._build_objective()

    def get_top_level_disc(self):
        return [self.mda]

    def _instantiate_mda(
        self, main_mda_class="MDAChain", sub_mda_class="MDAJacobi", **mda_options
    ):
        """Create MDA discipline"""
        if main_mda_class == "MDAChain":
            mda_options["sub_mda_class"] = sub_mda_class
        self.mda = self._mda_factory.create(
            main_mda_class, self.disciplines, **mda_options
        )

    @classmethod
    def get_sub_options_grammar(cls, **options):
        """
        When some options of the formulation depend on higher level
        options, a sub option schema may be specified here, mainly for
        use in the API

        :param options: options dict required to deduce the sub options grammar
        :returns: None, or the sub options grammar
        """
        main_mda = options.get("main_mda_class")
        if main_mda is None:
            raise ValueError(
                "main_mda_class option required \n"
                + "to deduce the sub options of MDF !"
            )
        factory = MDAFactory().factory
        return factory.get_options_grammar(main_mda)

    @classmethod
    def get_default_sub_options_values(cls, **options):
        """
        When some options of the formulation depend on higher level
        options, a sub option defaults may be specified here, mainly for
        use in the API

        :param options: options dict required to deduce the sub options grammar
        :returns: None, or the sub options defaults
        """
        main_mda = options.get("main_mda_class")
        if main_mda is None:
            raise ValueError(
                "main_mda_class option required \n"
                + "to deduce the sub options of MDF !"
            )
        factory = MDAFactory().factory
        return factory.get_default_options_values(main_mda)

    def _build_objective(self):
        """Builds the objective function on the MDA"""
        # Build the objective from the mda and the objective name
        self._build_objective_from_disc(self._objective_name, discipline=self.mda)

    def get_expected_workflow(self):
        return self.mda.get_expected_workflow()

    def get_expected_dataflow(self):
        return self.mda.get_expected_dataflow()

    def _update_design_space(self):
        """Update the design space by removing the coupling variables"""
        self._set_defaultinputs_from_ds()
        # No couplings in design space (managed by MDA)
        self._remove_couplings_from_ds()
        # Cleanup
        self._remove_unused_variables()

    def _remove_couplings_from_ds(self):
        """Removes the coupling variables from the design space"""
        design_space = self.opt_problem.design_space
        for coupling in self.mda.strong_couplings:
            if coupling in design_space.variables_names:
                design_space.remove_variable(coupling)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Gaussian mixture clustering algorithm
=====================================

The Gaussian mixture algorithm groups the data into clusters.
The number of clusters is fixed. Each cluster :math:`i=1, \\cdots, k` is
defined by a mean :math:`\\mu_i` and a covariance matrix :math:`\\Sigma_i`.

The prediction of the cluster value of a point is simply the cluster where the
probability density from the Gaussian distribution defined by the given mean
and covariance matrix is the highest:

.. math::

        \\operatorname{cluster}(x) =
            \\underset{i=1,\\cdots,k}{\\operatorname{argmax}}
            \\mathcal{N}(x; \\mu_i, \\Sigma_i) =
            \\underset{i=1,\\cdots,k}{\\operatorname{argmin}}
            \\|x-\\mu_i\\|_{\\Sigma_i^{-1}},

where :math:`\\mathcal{N}(x; \\mu_i, \\Sigma_i)` is the value of the
probability density function of a Gaussian random variable
:math:`X \\sim \\mathcal{N}(\\mu_i, \\Sigma_i)` at the point :math:`x` and
:math:`\\|x-\\mu_i\\|_{\\Sigma_i^{-1}} =
\\sqrt{(x-\\mu_i)^T \\Sigma_i^{-1} (x-\\mu_i)}`
is the Mahalanobis distance between :math:`x` and :math:`\\mu_i` weighted by
:math:`\\Sigma_i`.
Likewise, the probability of belonging to a cluster :math:`i=1, \\cdots, k` may
be determined through

.. math::

        \\mathbb{P}(x \\in C_i) = \\frac{\\mathcal{N}(x; \\mu_i, \\Sigma_i)}
            {\\sum_{j=1}^k \\mathcal{N}(x; \\mu_j, \\Sigma_j)},

where :math:`C_i = \\{x\\, | \\, \\operatorname{cluster}(x) = i \\}`.

When fitting the algorithm, the cluster centers :math:`\\mu_i` and the
covariance matrices :math:`\\Sigma_i` are computed using the
expectation-maximization algorithm.

This concept is implemented through the :class:`.GaussianMixture`
class which inherits from the :class:`.MLClusteringAlgo` class.

Dependence
----------
This clustering algorithm relies on the GaussianMixture class
of the `scikit-learn library <https://scikit-learn.org/stable/modules/
generated/sklearn.mixture.GaussianMixture.html>`_.
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library
from sklearn.mixture import GaussianMixture as SKLGaussianMixture

from gemseo.mlearning.cluster.cluster import MLClusteringAlgo

standard_library.install_aliases()


from gemseo import LOGGER


class GaussianMixture(MLClusteringAlgo):
    """ Gaussian mixture clustering algorithm. """

    ABBR = "GaussMix"

    def __init__(
        self, data, transformer=None, var_names=None, n_components=5, **parameters
    ):
        """Constructor.

        :param data: learning dataset.
        :type data: Dataset
        :param transformer: transformation strategy for data groups.
            If None, do not transform data. Default: None.
        :type transformer: dict(str)
        :param var_names: names of the variables to consider.
        :type var_names: list(str)
        :param n_components: number of Gaussian mixture components.
            Default: 5.
        :type n_components: int
        :param parameters: Scikit-learn algorithm parameters.
        """
        super(GaussianMixture, self).__init__(
            data,
            transformer=transformer,
            var_names=var_names,
            n_components=n_components,
            **parameters
        )
        self.algo = SKLGaussianMixture(n_components, **parameters)

    def _fit(self, data):
        """Fit the clustering model to the data and store labels.

        :param ndarray data: training data (2D).
        """
        self.algo.fit(data)
        self.labels = self.algo.predict(data)

    def _predict(self, data):
        """Predict cluster of data.

        :param ndarray data: data (2D).
        :return: clusters of data (1D).
        :rtype: ndarray(int).
        """
        return self.algo.predict(data)

    def _predict_proba_soft(self, data):
        """Predict probability of belonging to each cluster.

        :param ndarray data: data (2D).
        :return: probabilities for each cluster for each sample (2D). The sum
            of each row is one.
        :rtype: ndarray.
        """
        return self.algo.predict_proba(data)



# -*- coding: utf-8 -*-
# Copyright (c) 2009-2014, Luke Maurits <luke@maurits.id.au>
# All rights reserved.
# With contributions from:
# * Chris Clark
#  * Klein Stephane
#  * John Filleau
# PTable is forked from original Google Code page in April, 2015, and now
# maintained by Kane Blueriver <kxxoling@gmail.com>.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * The name of the author may not be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# Taken from :
# https://github.com/kxxoling/PTable
# https://pypi.python.org/pypi/PTable
"""
The pretty table core
*********************
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from builtins import int, map, range, str, zip
import copy
import math
import random
import re
import sys
import textwrap
import unicodedata

from future import standard_library

from ._compact import (basestring_, escape, itermap, str_types, uni_chr,
                       unicode_)


PY2 = sys.version_info.major == 2


standard_library.install_aliases()


# hrule styles
FRAME = 0
ALL = 1
NONE = 2
HEADER = 3

# Table styles
DEFAULT = 10
MSWORD_FRIENDLY = 11
PLAIN_COLUMNS = 12
RANDOM = 20

_re = re.compile("\033\[[0-9;]*m")


def _get_size(text):
    """

    :param text:

    """
    lines = text.split("\n")
    height = len(lines)
    width = max([_str_block_width(line) for line in lines])
    return width, height


class PrettyTable(object):

    def __init__(self, field_names=None, **kwargs):
        """

        :param encoding: Unicode encoding scheme used to decode any encoded input
        :param title: optional table title
        :param field_names: list or tuple of field names
        :param fields: list or tuple of field names to include in displays
        :param start: index of first data row to include in output
        :param end: index of last data row to include in output PLUS ONE
        :param header: print a header showing field names
        :param header_style: stylisation to apply to field names in header
        :param border: print a border around the table
        :param hrules: controls printing of horizontal rules after rows
        :param vrules: controls printing of vertical rules between columns
        :param int_format: controls formatting of integer data
        :param float_format: controls formatting of floating point data
        :param min_table_width: minimum desired table width
        :param max_table_width: maximum desired table width
        :param padding_width: number of spaces on either side of column data
        :param left_padding_width: number of spaces on left hand side of column data
        :param right_padding_width: number of spaces on right hand side of column data
        :param vertical_char: single character string used to draw vertical lines
        :param horizontal_char: single character string used to draw horizontal lines
        :param junction_char: single character string used to draw line junctions
        :param sortby: name of field to sort rows by
        :param sort_key: sorting key function
        :param valign: default valign for each row
        :param reversesort: True or False to sort in descending or ascending order
        :param oldsortslice: Slice rows before sorting in the

        """

        self.encoding = kwargs.get("encoding", "UTF-8")

        # Data
        self._field_names = []
        self._rows = []
        self.align = {}
        self.valign = {}
        self.max_width = {}
        self.min_width = {}
        self.int_format = {}
        self.float_format = {}
        if field_names:
            self.field_names = field_names
        else:
            self._widths = []

        # Options
        self._options = "title start end fields header border sortby reversesort sort_key attributes format hrules vrules".split()
        self._options.extend(
            "int_format float_format min_table_width max_table_width padding_width left_padding_width right_padding_width".split())
        self._options.extend(
            "vertical_char horizontal_char junction_char header_style valign xhtml print_empty oldsortslice".split())
        self._options.extend("align valign max_width min_width".split())
        for option in self._options:
            if option in kwargs:
                self._validate_option(option, kwargs[option])
            else:
                kwargs[option] = None

        self._title = kwargs["title"] or None
        self._start = kwargs["start"] or 0
        self._end = kwargs["end"] or None
        self._fields = kwargs["fields"] or None

        if kwargs["header"] in (True, False):
            self._header = kwargs["header"]
        else:
            self._header = True
        self._header_style = kwargs["header_style"] or None
        if kwargs["border"] in (True, False):
            self._border = kwargs["border"]
        else:
            self._border = True
        self._hrules = kwargs["hrules"] or FRAME
        self._vrules = kwargs["vrules"] or ALL

        self._sortby = kwargs["sortby"] or None
        if kwargs["reversesort"] in (True, False):
            self._reversesort = kwargs["reversesort"]
        else:
            self._reversesort = False
        self._sort_key = kwargs["sort_key"] or (lambda x: x)

        # Column specific arguments, use property.setters
        self.align = kwargs["align"] or {}
        self.valign = kwargs["valign"] or {}
        self.max_width = kwargs["max_width"] or {}
        self.min_width = kwargs["min_width"] or {}
        self.int_format = kwargs["int_format"] or {}
        self.float_format = kwargs["float_format"] or {}

        self._min_table_width = kwargs["min_table_width"] or None
        self._max_table_width = kwargs["max_table_width"] or None
        if kwargs["padding_width"] is None:
            self._padding_width = 1
        else:
            self._padding_width = kwargs["padding_width"]
        self._left_padding_width = kwargs["left_padding_width"] or None
        self._right_padding_width = kwargs["right_padding_width"] or None

        self._vertical_char = kwargs["vertical_char"] or self._unicode("|")
        self._horizontal_char = kwargs["horizontal_char"] or self._unicode("-")
        self._junction_char = kwargs["junction_char"] or self._unicode("+")

        if kwargs["print_empty"] in (True, False):
            self._print_empty = kwargs["print_empty"]
        else:
            self._print_empty = True
        if kwargs["oldsortslice"] in (True, False):
            self._oldsortslice = kwargs["oldsortslice"]
        else:
            self._oldsortslice = False
        self._format = kwargs["format"] or False
        self._xhtml = kwargs["xhtml"] or False
        self._attributes = kwargs["attributes"] or {}

    def _unicode(self, value):
        """

        :param value:

        """
        if not isinstance(value, basestring_):
            value = str(value)
        if not isinstance(value, unicode_):
            value = unicode_(value, self.encoding, "strict")
        return value

    def _justify(self, text, width, align):
        """

        :param text: param width:
        :param align:
        :param width:

        """
        excess = width - _str_block_width(text)
        if align == "l":
            return text + excess * " "
        elif align == "r":
            return excess * " " + text
        else:
            if excess % 2:
                # Uneven padding
                # Put more space on right if text is of odd length...
                if _str_block_width(text) % 2:
                    return (excess // 2) * " " + text + (excess // 2 + 1) * " "
                # and more space on left if text is of even length
                else:
                    return (excess // 2 + 1) * " " + text + (excess // 2) * " "
                    # Why distribute extra space this way?  To match the behaviour of
                    # the inbuilt str.center() method.
            else:
                # Equal padding on either side
                return (excess // 2) * " " + text + (excess // 2) * " "

    def __getattr__(self, name):

        if name == "rowcount":
            return len(self._rows)
        elif name == "colcount":
            if self._field_names:
                return len(self._field_names)
            elif self._rows:
                return len(self._rows[0])
            else:
                return 0
        else:
            raise AttributeError(name)

    def __getitem__(self, index):

        new = PrettyTable()
        new.field_names = self.field_names
        for attr in self._options:
            setattr(new, "_" + attr, getattr(self, "_" + attr))
        setattr(new, "_align", getattr(self, "_align"))
        if isinstance(index, slice):
            for row in self._rows[index]:
                new.add_row(row)
        elif isinstance(index, int):
            new.add_row(self._rows[index])
        else:
            raise Exception(
                "Index %s is invalid, must be an integer or slice" %
                str(index))
        return new

    def __str__(self):
        return self.__unicode__()

    def __unicode__(self):
        return self.get_string()

    ##############################
    # ATTRIBUTE VALIDATORS       #
    ##############################

    # The method _validate_option is all that should be used elsewhere in the code base to validate options.
    # It will call the appropriate validation method for that option.  The individual validation methods should
    # never need to be called directly (although nothing bad will happen if they *are*).
    # Validation happens in TWO places.
    # Firstly, in the property setters defined in the ATTRIBUTE MANAGMENT section.
    # Secondly, in the _get_options method, where keyword arguments are mixed
    # with persistent settings

    def _validate_option(self, option, val):
        """

        :param option: param val:
        :param val:

        """
        if option in ("field_names"):
            self._validate_field_names(val)
        elif option in ("start", "end", "max_width", "min_width", "min_table_width", "max_table_width", "padding_width",
                        "left_padding_width", "right_padding_width", "format"):
            self._validate_nonnegative_int(option, val)
        elif option in ("sortby"):
            self._validate_field_name(option, val)
        elif option in ("sort_key"):
            self._validate_function(option, val)
        elif option in ("hrules"):
            self._validate_hrules(option, val)
        elif option in ("vrules"):
            self._validate_vrules(option, val)
        elif option in ("fields"):
            self._validate_all_field_names(option, val)
        elif option in ("header", "border", "reversesort", "xhtml", "print_empty", "oldsortslice"):
            self._validate_true_or_false(option, val)
        elif option in ("header_style"):
            self._validate_header_style(val)
        elif option in ("int_format"):
            self._validate_int_format(option, val)
        elif option in ("float_format"):
            self._validate_float_format(option, val)
        elif option in ("vertical_char", "horizontal_char", "junction_char"):
            self._validate_single_char(option, val)
        elif option in ("attributes"):
            self._validate_attributes(option, val)

    def _validate_field_names(self, val):
        """

        :param val:

        """
        # Check for appropriate length
        if self._field_names:
            try:
                assert len(val) == len(self._field_names)
            except AssertionError:
                raise Exception("Field name list has incorrect number of values, (actual) %d!=%d (expected)" % (
                    len(val), len(self._field_names)))
        if self._rows:
            try:
                assert len(val) == len(self._rows[0])
            except AssertionError:
                raise Exception("Field name list has incorrect number of values, (actual) %d!=%d (expected)" % (
                    len(val), len(self._rows[0])))
        # Check for uniqueness
        try:
            assert len(val) == len(set(val))
        except AssertionError:
            raise Exception("Field names must be unique!")

    def _validate_header_style(self, val):
        """

        :param val:

        """
        try:
            assert val in ("cap", "title", "upper", "lower", None)
        except AssertionError:
            raise Exception(
                "Invalid header style, use cap, title, upper, lower or None!")

    def _validate_align(self, val):
        """

        :param val:

        """
        try:
            assert val in ["l", "c", "r"]
        except AssertionError:
            raise Exception("Alignment %s is invalid, use l, c or r!" % val)

    def _validate_valign(self, val):
        """

        :param val:

        """
        try:
            assert val in ["t", "m", "b", None]
        except AssertionError:
            raise Exception(
                "Alignment %s is invalid, use t, m, b or None!" %
                val)

    def _validate_nonnegative_int(self, name, val):
        """

        :param name: param val:
        :param val:

        """
        try:
            assert int(val) >= 0
        except AssertionError:
            raise Exception(
                "Invalid value for %s: %s!" %
                (name, self._unicode(val)))

    def _validate_true_or_false(self, name, val):
        """

        :param name: param val:
        :param val:

        """
        try:
            assert val in (True, False)
        except AssertionError:
            raise Exception(
                "Invalid value for %s!  Must be True or False." %
                name)

    def _validate_int_format(self, name, val):
        """

        :param name: param val:
        :param val:

        """
        if val == "":
            return
        try:
            assert type(val) in str_types
            assert val.isdigit()
        except AssertionError:
            raise Exception(
                "Invalid value for %s!  Must be an integer format string." %
                name)

    def _validate_float_format(self, name, val):
        """

        :param name: param val:
        :param val:

        """
        return
        if val == "":
            return
        try:
            val = val.rsplit('f')[0]
            assert type(val) in str_types
            assert "." in val
            bits = val.split(".")
            assert len(bits) <= 2
            assert bits[0] == "" or bits[0].isdigit()
            assert bits[1] == "" or bits[1].isdigit()
        except AssertionError:
            raise Exception(
                "Invalid value for %s!  Must be a float format string." %
                name)

    def _validate_function(self, name, val):
        """

        :param name: param val:
        :param val:

        """
        try:
            assert hasattr(val, "__call__")
        except AssertionError:
            raise Exception(
                "Invalid value for %s!  Must be a function." %
                name)

    def _validate_hrules(self, name, val):
        """

        :param name: param val:
        :param val:

        """
        try:
            assert val in (ALL, FRAME, HEADER, NONE)
        except AssertionError:
            raise Exception(
                "Invalid value for %s!  Must be ALL, FRAME, HEADER or NONE." %
                name)

    def _validate_vrules(self, name, val):
        """

        :param name: param val:
        :param val:

        """
        try:
            assert val in (ALL, FRAME, NONE)
        except AssertionError:
            raise Exception(
                "Invalid value for %s!  Must be ALL, FRAME, or NONE." %
                name)

    def _validate_field_name(self, name, val):
        """

        :param name: param val:
        :param val:

        """
        try:
            assert (val in self._field_names) or (val is None)
        except AssertionError:
            raise Exception("Invalid field name: %s!" % val)

    def _validate_all_field_names(self, name, val):
        """

        :param name: param val:
        :param val:

        """
        try:
            for x in val:
                self._validate_field_name(name, x)
        except AssertionError:
            raise Exception("fields must be a sequence of field names!")

    def _validate_single_char(self, name, val):
        """

        :param name: param val:
        :param val:

        """
        try:
            assert _str_block_width(val) == 1
        except AssertionError:
            raise Exception(
                "Invalid value for %s!  Must be a string of length 1." %
                name)

    def _validate_attributes(self, name, val):
        """

        :param name: param val:
        :param val:

        """
        try:
            assert isinstance(val, dict)
        except AssertionError:
            raise Exception(
                "attributes must be a dictionary of name/value pairs!")

    ##############################
    # ATTRIBUTE MANAGEMENT       #
    ##############################

    @property
    def field_names(self):
        """List or tuple of field names"""
        return self._field_names

    @field_names.setter
    def field_names(self, val):
        """

        :param val:

        """
        val = [self._unicode(x) for x in val]
        self._validate_option("field_names", val)
        if self._field_names:
            old_names = self._field_names[:]
        self._field_names = val
        if self._align and old_names:
            for old_name, new_name in zip(old_names, val):
                self._align[new_name] = self._align[old_name]
            for old_name in old_names:
                if old_name not in self._align:
                    self._align.pop(old_name)
        else:
            self.align = "c"
        if self._valign and old_names:
            for old_name, new_name in zip(old_names, val):
                self._valign[new_name] = self._valign[old_name]
            for old_name in old_names:
                if old_name not in self._valign:
                    self._valign.pop(old_name)
        else:
            self.valign = "t"

    @property
    def align(self):
        """Controls alignment of fields

        :param align: alignment

        """
        return self._align

    @align.setter
    def align(self, val):
        """

        :param val:

        """
        if not self._field_names:
            self._align = {}
        elif val is None or (isinstance(val, dict) and len(val) == 0):
            for field in self._field_names:
                self._align[field] = "c"
        else:
            self._validate_align(val)
            for field in self._field_names:
                self._align[field] = val

    @property
    def valign(self):
        """Controls vertical alignment of fields

        :param valign: vertical alignment

        """
        return self._valign

    @valign.setter
    def valign(self, val):
        """

        :param val:

        """
        if not self._field_names:
            self._valign = {}
        elif val is None or (isinstance(val, dict) and len(val) == 0):
            for field in self._field_names:
                self._valign[field] = "t"
        else:
            self._validate_valign(val)
            for field in self._field_names:
                self._valign[field] = val

    @property
    def max_width(self):
        """Controls maximum width of fields

        :param max_width: maximum width integer

        """
        return self._max_width

    @max_width.setter
    def max_width(self, val):
        """

        :param val:

        """
        if val is None or (isinstance(val, dict) and len(val) == 0):
            self._max_width = {}
        else:
            self._validate_option("max_width", val)
            for field in self._field_names:
                self._max_width[field] = val

    @property
    def min_width(self):
        """Controls minimum width of fields

        :param min_width: minimum width integer

        """
        if self.header:
            fields = self._field_names
        else:
            fields = self._rows[0] if self._rows else []
        result = {
            # minimum column width can't be lesser
            # than header's length
            name: max(
                _str_block_width(unicode_(name)),
                self._min_width.get(name, 0)
            )
            for name in fields
        }
        return result

    @min_width.setter
    def min_width(self, val):
        """

        :param val:

        """
        if val is None or (isinstance(val, dict) and len(val) == 0):
            self._min_width = {}
        else:
            self._validate_option("min_width", val)
            for field in self._field_names:
                self._min_width[field] = val

    @property
    def min_table_width(self):
        """ """
        return self._min_table_width

    @min_table_width.setter
    def min_table_width(self, val):
        """

        :param val:

        """
        self._validate_option("min_table_width", val)
        self._min_table_width = val

    @property
    def max_table_width(self):
        """ """
        return self._max_table_width

    @max_table_width.setter
    def max_table_width(self, val):
        """

        :param val:

        """
        self._validate_option("max_table_width", val)
        self._max_table_width = val

    @property
    def fields(self):
        """List or tuple of field names to include in displays"""
        return self._fields

    @fields.setter
    def fields(self, val):
        """

        :param val:

        """
        self._validate_option("fields", val)
        self._fields = val

    @property
    def title(self):
        """Optional table title

        :param title: table title

        """
        return self._title

    @title.setter
    def title(self, val):
        """

        :param val:

        """
        self._title = self._unicode(val)

    @property
    def start(self):
        """Start index of the range of rows to print

        :param start: index of first data row to include in output

        """
        return self._start

    @start.setter
    def start(self, val):
        """

        :param val:

        """
        self._validate_option("start", val)
        self._start = val

    @property
    def end(self):
        """End index of the range of rows to print

        :param end: index of last data row to include in output PLUS ONE

        """
        return self._end

    @end.setter
    def end(self, val):
        """

        :param val:

        """
        self._validate_option("end", val)
        self._end = val

    @property
    def sortby(self):
        """Name of field by which to sort rows

        :param sortby: field name to sort by

        """
        return self._sortby

    @sortby.setter
    def sortby(self, val):
        """

        :param val:

        """
        self._validate_option("sortby", val)
        self._sortby = val

    @property
    def reversesort(self):
        """Controls direction of sorting (ascending vs descending)

        :param reveresort: set to True to sort by descending order

        """
        return self._reversesort

    @reversesort.setter
    def reversesort(self, val):
        """

        :param val:

        """
        self._validate_option("reversesort", val)
        self._reversesort = val

    @property
    def sort_key(self):
        """Sorting key function, applied to data points before sorting

        :param sort_key: a function which takes one argument and returns something to be sorted

        """
        return self._sort_key

    @sort_key.setter
    def sort_key(self, val):
        """

        :param val:

        """
        self._validate_option("sort_key", val)
        self._sort_key = val

    @property
    def header(self):
        """Controls printing of table header with field names

        :param header: print a header showing field names

        """
        return self._header

    @header.setter
    def header(self, val):
        """

        :param val:

        """
        self._validate_option("header", val)
        self._header = val

    @property
    def header_style(self):
        """Controls stylisation applied to field names in header

        :param header_style: stylisation to apply to field names in header

        """
        return self._header_style

    @header_style.setter
    def header_style(self, val):
        """

        :param val:

        """
        self._validate_header_style(val)
        self._header_style = val

    @property
    def border(self):
        """Controls printing of border around table

        :param border: print a border around the table

        """
        return self._border

    @border.setter
    def border(self, val):
        """

        :param val:

        """
        self._validate_option("border", val)
        self._border = val

    @property
    def hrules(self):
        """Controls printing of horizontal rules after rows

        :param hrules: horizontal rules style

        """
        return self._hrules

    @hrules.setter
    def hrules(self, val):
        """

        :param val:

        """
        self._validate_option("hrules", val)
        self._hrules = val

    @property
    def vrules(self):
        """Controls printing of vertical rules between columns

        :param vrules: vertical rules style

        """
        return self._vrules

    @vrules.setter
    def vrules(self, val):
        """

        :param val:

        """
        self._validate_option("vrules", val)
        self._vrules = val

    @property
    def int_format(self):
        """Controls formatting of integer data

        :param int_format: integer format string

        """
        return self._int_format

    @int_format.setter
    def int_format(self, val):
        """

        :param val:

        """
        if val is None or (isinstance(val, dict) and len(val) == 0):
            self._int_format = {}
        else:
            self._validate_option("int_format", val)
            for field in self._field_names:
                self._int_format[field] = val

    @property
    def float_format(self):
        """Controls formatting of floating point data

        :param float_format: floating point format string

        """
        return self._float_format

    @float_format.setter
    def float_format(self, val):
        """

        :param val:

        """
        if val is None or (isinstance(val, dict) and len(val) == 0):
            self._float_format = {}
        else:
            self._validate_option("float_format", val)
            for field in self._field_names:
                self._float_format[field] = val

    @property
    def padding_width(self):
        """The number of empty spaces between a column's edge and its content

        :param padding_width: number of spaces

        """
        return self._padding_width

    @padding_width.setter
    def padding_width(self, val):
        """

        :param val:

        """
        self._validate_option("padding_width", val)
        self._padding_width = val

    @property
    def left_padding_width(self):
        """The number of empty spaces between a column's left edge and its content

        :param left_padding: number of spaces

        """
        return self._left_padding_width

    @left_padding_width.setter
    def left_padding_width(self, val):
        """

        :param val:

        """
        self._validate_option("left_padding_width", val)
        self._left_padding_width = val

    @property
    def right_padding_width(self):
        """The number of empty spaces between a column's right edge and its content

        :param right_padding: number of spaces

        """
        return self._right_padding_width

    @right_padding_width.setter
    def right_padding_width(self, val):
        """

        :param val:

        """
        self._validate_option("right_padding_width", val)
        self._right_padding_width = val

    @property
    def vertical_char(self):
        """The charcter used when printing table borders to draw vertical lines

        :param vertical_char: single character string used to draw vertical lines

        """
        return self._vertical_char

    @vertical_char.setter
    def vertical_char(self, val):
        """

        :param val:

        """
        val = self._unicode(val)
        self._validate_option("vertical_char", val)
        self._vertical_char = val

    @property
    def horizontal_char(self):
        """The charcter used when printing table borders to draw horizontal lines

        :param horizontal_char: single character string used to draw horizontal lines

        """
        return self._horizontal_char

    @horizontal_char.setter
    def horizontal_char(self, val):
        """

        :param val:

        """
        val = self._unicode(val)
        self._validate_option("horizontal_char", val)
        self._horizontal_char = val

    @property
    def junction_char(self):
        """The charcter used when printing table borders to draw line junctions

        :param junction_char: single character string used to draw line junctions

        """
        return self._junction_char

    @junction_char.setter
    def junction_char(self, val):
        """

        :param val:

        """
        val = self._unicode(val)
        self._validate_option("vertical_char", val)
        self._junction_char = val

    @property
    def format(self):
        """Controls whether or not HTML tables are formatted to match styling options

        :param format: True or False

        """
        return self._format

    @format.setter
    def format(self, val):
        """

        :param val:

        """
        self._validate_option("format", val)
        self._format = val

    @property
    def print_empty(self):
        """Controls whether or not empty tables produce a header and frame or just an empty string

        :param print_empty: True or False

        """
        return self._print_empty

    @print_empty.setter
    def print_empty(self, val):
        """

        :param val:

        """
        self._validate_option("print_empty", val)
        self._print_empty = val

    @property
    def attributes(self):
        """A dictionary of HTML attribute name/value pairs to be included in the <table> tag when printing HTML

        Arguments:

        attributes - dictionary of attributes


        """
        return self._attributes

    @attributes.setter
    def attributes(self, val):
        """

        :param val:

        """
        self._validate_option("attributes", val)
        self._attributes = val

    @property
    def oldsortslice(self):
        """oldsortslice - Slice rows before sorting in the 'old style'
        """
        return self._oldsortslice

    @oldsortslice.setter
    def oldsortslice(self, val):
        """

        :param val:

        """
        self._validate_option("oldsortslice", val)
        self._oldsortslice = val

    ##############################
    # OPTION MIXER               #
    ##############################

    def _get_options(self, kwargs):
        """

        :param kwargs:

        """

        options = {}
        for option in self._options:
            if option in kwargs:
                self._validate_option(option, kwargs[option])
                options[option] = kwargs[option]
            else:
                options[option] = getattr(self, "_" + option)
        return options

    ##############################
    # PRESET STYLE LOGIC         #
    ##############################

    def set_style(self, style):
        """

        :param style:

        """

        if style == DEFAULT:
            self._set_default_style()
        elif style == MSWORD_FRIENDLY:
            self._set_msword_style()
        elif style == PLAIN_COLUMNS:
            self._set_columns_style()
        elif style == RANDOM:
            self._set_random_style()
        else:
            raise Exception("Invalid pre-set style!")

    def _set_default_style(self):
        """ """

        self.header = True
        self.border = True
        self._hrules = FRAME
        self._vrules = ALL
        self.padding_width = 1
        self.left_padding_width = 1
        self.right_padding_width = 1
        self.vertical_char = "|"
        self.horizontal_char = "-"
        self.junction_char = "+"

    def _set_msword_style(self):
        """ """

        self.header = True
        self.border = True
        self._hrules = NONE
        self.padding_width = 1
        self.left_padding_width = 1
        self.right_padding_width = 1
        self.vertical_char = "|"

    def _set_columns_style(self):
        """ """

        self.header = True
        self.border = False
        self.padding_width = 1
        self.left_padding_width = 0
        self.right_padding_width = 8

    def _set_random_style(self):
        """ """

        # Just for fun!
        self.header = random.choice((True, False))
        self.border = random.choice((True, False))
        self._hrules = random.choice((ALL, FRAME, HEADER, NONE))
        self._vrules = random.choice((ALL, FRAME, NONE))
        self.left_padding_width = random.randint(0, 5)
        self.right_padding_width = random.randint(0, 5)
        self.vertical_char = random.choice("~!@#$%^&*()_+|-=\{}[];':\",./;<>?")
        self.horizontal_char = random.choice(
            "~!@#$%^&*()_+|-=\{}[];':\",./;<>?")
        self.junction_char = random.choice("~!@#$%^&*()_+|-=\{}[];':\",./;<>?")

    ##############################
    # DATA INPUT METHODS         #
    ##############################

    def add_row(self, row):
        """Add a row to the table

        :param row: row of data
        :param has: fields

        """

        if self._field_names and len(row) != len(self._field_names):
            raise Exception(
                "Row has incorrect number of values, (actual) %d!=%d (expected)" % (len(row), len(self._field_names)))
        if not self._field_names:
            self.field_names = [
                ("Field %d" % (n + 1)) for n in range(0, len(row))]
        self._rows.append(list(row))

    def del_row(self, row_index):
        """Delete a row to the table

        :param row_index: The index of the row you want to delete

        """

        if row_index > len(self._rows) - 1:
            raise Exception(
                "Cant delete row at index %d, table only has %d rows!" %
                (row_index, len(
                    self._rows)))
        del self._rows[row_index]

    def add_column(self, fieldname, column, align="c", valign="t"):
        """Add a column to the table.

        :param fieldname: name of the field to contain the new column of data
        :param column: column of data
        :param table: has rows
        :param align: desired alignment for this column (Default value = "c")
        :param valign: desired vertical alignment for new columns (Default value = "t")

        """

        if len(self._rows) in (0, len(column)):
            self._validate_align(align)
            self._validate_valign(valign)
            self._field_names.append(fieldname)
            self._align[fieldname] = align
            self._valign[fieldname] = valign
            for i in range(0, len(column)):
                if len(self._rows) < i + 1:
                    self._rows.append([])
                self._rows[i].append(column[i])
        else:
            raise Exception(
                "Column length %d does not match number of rows %d!" %
                (len(column), len(
                    self._rows)))

    def clear_rows(self):
        """Delete all rows from the table but keep the current field names"""

        self._rows = []

    def clear(self):
        """Delete all rows and field names from the table, maintaining nothing but styling options"""

        self._rows = []
        self._field_names = []
        self._widths = []

    ##############################
    # MISC PUBLIC METHODS        #
    ##############################

    def copy(self):
        """ """
        return copy.deepcopy(self)

    ##############################
    # MISC PRIVATE METHODS       #
    ##############################

    def _format_value(self, field, value):
        """

        :param field: param value:
        :param value:

        """
        if isinstance(value, int) and field in self._int_format:
            value = self._unicode(("%%%sd" % self._int_format[field]) % value)
        elif isinstance(value, float) and field in self._float_format:
            value = self._unicode(
                (self._float_format[field]) %
                value)
        return self._unicode(value)

    def _compute_table_width(self, options):
        """

        :param options:

        """
        table_width = 2 if options["vrules"] in (FRAME, ALL) else 0
        per_col_padding = sum(self._get_padding_widths(options))
        for index, fieldname in enumerate(self.field_names):
            if not options["fields"] or (
                    options["fields"] and fieldname in options["fields"]):
                table_width += self._widths[index] + per_col_padding
        return table_width

    def _compute_widths(self, rows, options):
        """

        :param rows: param options:
        :param options:

        """
        if options["header"]:
            widths = [_get_size(field)[0] for field in self._field_names]
        else:
            widths = len(self.field_names) * [0]

        for row in rows:
            for index, value in enumerate(row):
                fieldname = self.field_names[index]
                if fieldname in self.max_width:
                    widths[index] = max(
                        widths[index], min(
                            _get_size(value)[0], self.max_width[fieldname]))
                else:
                    widths[index] = max(widths[index], _get_size(value)[0])
                if fieldname in self.min_width:
                    widths[index] = max(
                        widths[index],
                        self.min_width[fieldname])
        self._widths = widths

        # Are we exceeding max_table_width?
        if self._max_table_width:
            table_width = self._compute_table_width(options)
            if table_width > self._max_table_width:
                # get dict with minimum widths for fields
                min_width = self.min_width

                # first calculate width for paddings and vrules this
                # space we can't shrink.

                # Space for vrules
                nonshrinkable = 2 if options["vrules"] in (FRAME, ALL) else 0
                # Space for vrules between columns
                nonshrinkable += len(self._field_names) - 1
                # Space for padding in each column
                per_col_padding = sum(self._get_padding_widths(options))
                nonshrinkable += len(widths) * per_col_padding
                # Min space for each column
                nonshrinkable += sum(min_width.values())

                # Shrink widths in proportion
                scale = float(
                    self._max_table_width - nonshrinkable) / (table_width -
                                                              nonshrinkable)

                def calculate_new_width(field_name, old_width):
                    """

                    :param field_name: param old_width:
                    :param old_width:

                    """
                    width = min_width[field_name]
                    # scale according to recalculated table width
                    scaled_part = int(math.floor((old_width - width) * scale))
                    # enforce minimum column width as 1 symbol
                    return max(1, width + scaled_part)

                widths = list(map(
                    calculate_new_width,
                    self._field_names,
                    widths
                ))
                self._widths = widths

        # Are we under min_table_width or title width?
        if self._min_table_width or options["title"]:
            if options["title"]:
                title_width = len(
                    options["title"]) + sum(self._get_padding_widths(options))
                if options["vrules"] in (FRAME, ALL):
                    title_width += 2
            else:
                title_width = 0
            min_table_width = self.min_table_width or 0
            min_width = max(title_width, min_table_width)
            table_width = self._compute_table_width(options)
            if table_width < min_width:
                # Grow widths in proportion
                scale = 1.0 * min_width / table_width
                widths = [int(math.ceil(w * scale)) for w in widths]
                self._widths = widths

    def _get_padding_widths(self, options):
        """

        :param options:

        """

        if options["left_padding_width"] is not None:
            lpad = options["left_padding_width"]
        else:
            lpad = options["padding_width"]
        if options["right_padding_width"] is not None:
            rpad = options["right_padding_width"]
        else:
            rpad = options["padding_width"]
        return lpad, rpad

    def _get_rows(self, options):
        """

        :param options: dictionary of option settings

        """

        if options["oldsortslice"]:
            rows = copy.deepcopy(self._rows[options["start"]:options["end"]])
        else:
            rows = copy.deepcopy(self._rows)

        # Sort
        if options["sortby"]:
            sortindex = self._field_names.index(options["sortby"])
            # Decorate
            rows = [[row[sortindex]] + row for row in rows]
            # Sort
            rows.sort(reverse=options["reversesort"], key=options["sort_key"])
            # Undecorate
            rows = [row[1:] for row in rows]

        # Slice if necessary
        if not options["oldsortslice"]:
            rows = rows[options["start"]:options["end"]]

        return rows

    def _format_row(self, row, options):
        """

        :param row: param options:
        :param options:

        """
        return [self._format_value(field, value)
                for (field, value) in zip(self._field_names, row)]

    def _format_rows(self, rows, options):
        """

        :param rows: param options:
        :param options:

        """
        return [self._format_row(row, options) for row in rows]

    ##############################
    # PLAIN TEXT STRING METHODS  #
    ##############################

    def get_string(self, **kwargs):
        """

        :param title: optional table title
        :param start: index of first data row to include in output
        :param end: index of last data row to include in output PLUS ONE
        :param fields: names of fields
        :param header: print a header showing field names
        :param border: print a border around the table
        :param hrules: controls printing of horizontal rules after rows
        :param vrules: controls printing of vertical rules between columns
        :param int_format: controls formatting of integer data
        :param float_format: controls formatting of floating point data
        :param padding_width: number of spaces on either side of column data
        :param left_padding_width: number of spaces on left hand side of column data
        :param right_padding_width: number of spaces on right hand side of column data
        :param vertical_char: single character string used to draw vertical lines
        :param horizontal_char: single character string used to draw horizontal lines
        :param junction_char: single character string used to draw line junctions
        :param sortby: name of field to sort rows by
        :param sort_key: sorting key function
        :param reversesort: True or False to sort in descending or ascending order
        :param print: empty

        """

        options = self._get_options(kwargs)

        lines = []

        # Don't think too hard about an empty table
        # Is this the desired behaviour?  Maybe we should still print the
        # header?
        if self.rowcount == 0 and (
                not options["print_empty"] or not options["border"]):
            return ""

        # Get the rows we need to print, taking into account slicing, sorting,
        # etc.
        rows = self._get_rows(options)

        # Turn all data in all rows into Unicode, formatted as desired
        formatted_rows = self._format_rows(rows, options)

        # Compute column widths
        self._compute_widths(formatted_rows, options)
        self._hrule = self._stringify_hrule(options)

        # Add title
        title = options["title"] or self._title
        if title:
            lines.append(self._stringify_title(title, options))

        # Add header or top of border
        if options["header"]:
            lines.append(self._stringify_header(options))
        elif options["border"] and options["hrules"] in (ALL, FRAME):
            lines.append(self._hrule)

        # Add rows
        for row in formatted_rows:
            lines.append(self._stringify_row(row, options))

        # Add bottom of border
        if options["border"] and options["hrules"] == FRAME:
            lines.append(self._hrule)

        return self._unicode("\n").join(lines)

    def _stringify_hrule(self, options):
        """

        :param options:

        """

        if not options["border"]:
            return ""
        lpad, rpad = self._get_padding_widths(options)
        if options['vrules'] in (ALL, FRAME):
            bits = [options["junction_char"]]
        else:
            bits = [options["horizontal_char"]]
        # For tables with no data or fieldnames
        if not self._field_names:
            bits.append(options["junction_char"])
            return "".join(bits)
        for field, width in zip(self._field_names, self._widths):
            if options["fields"] and field not in options["fields"]:
                continue
            bits.append((width + lpad + rpad) * options["horizontal_char"])
            if options['vrules'] == ALL:
                bits.append(options["junction_char"])
            else:
                bits.append(options["horizontal_char"])
        if options["vrules"] == FRAME:
            bits.pop()
            bits.append(options["junction_char"])
        return "".join(bits)

    def _stringify_title(self, title, options):
        """

        :param title: param options:
        :param options:

        """

        lines = []
        lpad, rpad = self._get_padding_widths(options)
        if options["border"]:
            if options["vrules"] == ALL:
                options["vrules"] = FRAME
                lines.append(self._stringify_hrule(options))
                options["vrules"] = ALL
            elif options["vrules"] == FRAME:
                lines.append(self._stringify_hrule(options))
        bits = []
        endpoint = options["vertical_char"] if options[
            "vrules"] in (ALL, FRAME) else " "
        bits.append(endpoint)
        title = " " * lpad + title + " " * rpad
        bits.append(self._justify(title, len(self._hrule) - 2, "c"))
        bits.append(endpoint)
        lines.append("".join(bits))
        return "\n".join(lines)

    def _stringify_header(self, options):
        """

        :param options:

        """

        bits = []
        lpad, rpad = self._get_padding_widths(options)
        if options["border"]:
            if options["hrules"] in (ALL, FRAME):
                bits.append(self._hrule)
                bits.append("\n")
            if options["vrules"] in (ALL, FRAME):
                bits.append(options["vertical_char"])
            else:
                bits.append(" ")
        # For tables with no data or field names
        if not self._field_names:
            if options["vrules"] in (ALL, FRAME):
                bits.append(options["vertical_char"])
            else:
                bits.append(" ")
        for field, width, in zip(self._field_names, self._widths):
            if options["fields"] and field not in options["fields"]:
                continue
            if self._header_style == "cap":
                fieldname = field.capitalize()
            elif self._header_style == "title":
                fieldname = field.title()
            elif self._header_style == "upper":
                fieldname = field.upper()
            elif self._header_style == "lower":
                fieldname = field.lower()
            else:
                fieldname = field
            bits.append(
                " " *
                lpad +
                self._justify(
                    fieldname,
                    width,
                    self._align[field]) +
                " " *
                rpad)
            if options["border"]:
                if options["vrules"] == ALL:
                    bits.append(options["vertical_char"])
                else:
                    bits.append(" ")
        # If vrules is FRAME, then we just appended a space at the end
        # of the last field, when we really want a vertical character
        if options["border"] and options["vrules"] == FRAME:
            bits.pop()
            bits.append(options["vertical_char"])
        if options["border"] and options["hrules"] != NONE:
            bits.append("\n")
            bits.append(self._hrule)
        return "".join(bits)

    def _stringify_row(self, row, options):
        """

        :param row: param options:
        :param options:

        """

        for index, field, value, width, in zip(
                list(range(0, len(row))), self._field_names, row, self._widths):
            # Enforce max widths
            lines = value.split("\n")
            new_lines = []
            for line in lines:
                if _str_block_width(line) > width:
                    line = textwrap.fill(line, width)
                new_lines.append(line)
            lines = new_lines
            value = "\n".join(lines)
            row[index] = value

        row_height = 0
        for c in row:
            h = _get_size(c)[1]
            if h > row_height:
                row_height = h

        bits = []
        lpad, rpad = self._get_padding_widths(options)
        for y in range(0, row_height):
            bits.append([])
            if options["border"]:
                if options["vrules"] in (ALL, FRAME):
                    bits[y].append(self.vertical_char)
                else:
                    bits[y].append(" ")

        for field, value, width, in zip(self._field_names, row, self._widths):

            valign = self._valign[field]
            lines = value.split("\n")
            dHeight = row_height - len(lines)
            if dHeight:
                if valign == "m":
                    lines = [
                        ""] * int(dHeight / 2) + lines + [""] * (dHeight - int(dHeight / 2))
                elif valign == "b":
                    lines = [""] * dHeight + lines
                else:
                    lines = lines + [""] * dHeight

            y = 0
            for l in lines:
                if options["fields"] and field not in options["fields"]:
                    continue

                bits[y].append(
                    " " *
                    lpad +
                    self._justify(
                        l,
                        width,
                        self._align[field]) +
                    " " *
                    rpad)
                if options["border"]:
                    if options["vrules"] == ALL:
                        bits[y].append(self.vertical_char)
                    else:
                        bits[y].append(" ")
                y += 1

        # If vrules is FRAME, then we just appended a space at the end
        # of the last field, when we really want a vertical character
        for y in range(0, row_height):
            if options["border"] and options["vrules"] == FRAME:
                bits[y].pop()
                bits[y].append(options["vertical_char"])

        if options["border"] and options["hrules"] == ALL:
            bits[row_height - 1].append("\n")
            bits[row_height - 1].append(self._hrule)

        for y in range(0, row_height):
            bits[y] = "".join(bits[y])

        return "\n".join(bits)

    def paginate(self, page_length=58, **kwargs):
        """

        :param page_length: Default value = 58)

        """

        pages = []
        kwargs["start"] = kwargs.get("start", 0)
        true_end = kwargs.get("end", self.rowcount)
        while True:
            kwargs["end"] = min(kwargs["start"] + page_length, true_end)
            pages.append(self.get_string(**kwargs))
            if kwargs["end"] == true_end:
                break
            kwargs["start"] += page_length
        return "\f".join(pages)

    ##############################
    # HTML STRING METHODS        #
    ##############################

    def get_html_string(self, **kwargs):
        """

        :param title: optional table title
        :param start: index of first data row to include in output
        :param end: index of last data row to include in output PLUS ONE
        :param fields: names of fields
        :param header: print a header showing field names
        :param border: print a border around the table
        :param hrules: controls printing of horizontal rules after rows
        :param vrules: controls printing of vertical rules between columns
        :param int_format: controls formatting of integer data
        :param float_format: controls formatting of floating point data
        :param padding_width: number of spaces on either side of column data
        :param left_padding_width: number of spaces on left hand side of column data
        :param right_padding_width: number of spaces on right hand side of column data
        :param sortby: name of field to sort rows by
        :param sort_key: sorting key function
        :param attributes: dictionary of name
        :param xhtml: print

        """

        options = self._get_options(kwargs)

        if options["format"]:
            string = self._get_formatted_html_string(options)
        else:
            string = self._get_simple_html_string(options)

        return string

    def _get_simple_html_string(self, options):
        """

        :param options:

        """

        lines = []
        if options["xhtml"]:
            linebreak = "<br/>"
        else:
            linebreak = "<br>"

        open_tag = list()
        open_tag.append("<table")
        if options["attributes"]:
            for attr_name in options["attributes"]:
                open_tag.append(
                    " %s=\"%s\"" %
                    (attr_name, options["attributes"][attr_name]))
        open_tag.append(">")
        lines.append("".join(open_tag))

        # Title
        title = options["title"] or self._title
        if title:
            cols = len(
                options["fields"]) if options["fields"] else len(
                self.field_names)
            lines.append("    <tr>")
            lines.append("        <td colspan=%d>%s</td>" % (cols, title))
            lines.append("    </tr>")

        # Headers
        if options["header"]:
            lines.append("    <tr>")
            for field in self._field_names:
                if options["fields"] and field not in options["fields"]:
                    continue
                lines.append(
                    "        <th>%s</th>" %
                    escape(field).replace(
                        "\n",
                        linebreak))
            lines.append("    </tr>")

        # Data
        rows = self._get_rows(options)
        formatted_rows = self._format_rows(rows, options)
        for row in formatted_rows:
            lines.append("    <tr>")
            for field, datum in zip(self._field_names, row):
                if options["fields"] and field not in options["fields"]:
                    continue
                lines.append(
                    "        <td>%s</td>" %
                    escape(datum).replace(
                        "\n",
                        linebreak))
            lines.append("    </tr>")

        lines.append("</table>")

        return self._unicode("\n").join(lines)

    def _get_formatted_html_string(self, options):
        """

        :param options:

        """

        lines = []
        lpad, rpad = self._get_padding_widths(options)
        if options["xhtml"]:
            linebreak = "<br/>"
        else:
            linebreak = "<br>"

        open_tag = list()
        open_tag.append("<table")
        if options["border"]:
            if options["hrules"] == ALL and options["vrules"] == ALL:
                open_tag.append(" frame=\"box\" rules=\"all\"")
            elif options["hrules"] == FRAME and options["vrules"] == FRAME:
                open_tag.append(" frame=\"box\"")
            elif options["hrules"] == FRAME and options["vrules"] == ALL:
                open_tag.append(" frame=\"box\" rules=\"cols\"")
            elif options["hrules"] == FRAME:
                open_tag.append(" frame=\"hsides\"")
            elif options["hrules"] == ALL:
                open_tag.append(" frame=\"hsides\" rules=\"rows\"")
            elif options["vrules"] == FRAME:
                open_tag.append(" frame=\"vsides\"")
            elif options["vrules"] == ALL:
                open_tag.append(" frame=\"vsides\" rules=\"cols\"")
        if options["attributes"]:
            for attr_name in options["attributes"]:
                open_tag.append(
                    " %s=\"%s\"" %
                    (attr_name, options["attributes"][attr_name]))
        open_tag.append(">")
        lines.append("".join(open_tag))

        # Title
        title = options["title"] or self._title
        if title:
            cols = len(
                options["fields"]) if options["fields"] else len(
                self.field_names)
            lines.append("    <tr>")
            lines.append("        <td colspan=%d>%s</td>" % (cols, title))
            lines.append("    </tr>")

        # Headers
        if options["header"]:
            lines.append("    <tr>")
            for field in self._field_names:
                if options["fields"] and field not in options["fields"]:
                    continue
                lines.append(
                    "        <th style=\"padding-left: %dem; padding-right: %dem; text-align: center\">%s</th>" % (
                        lpad, rpad, escape(field).replace("\n", linebreak)))
            lines.append("    </tr>")

        # Data
        rows = self._get_rows(options)
        formatted_rows = self._format_rows(rows, options)
        aligns = []
        valigns = []
        for field in self._field_names:
            aligns.append(
                {"l": "left", "r": "right", "c": "center"}[self._align[field]])
            valigns.append(
                {"t": "top", "m": "middle", "b": "bottom"}[self._valign[field]])
        for row in formatted_rows:
            lines.append("    <tr>")
            for field, datum, align, valign in zip(
                    self._field_names, row, aligns, valigns):
                if options["fields"] and field not in options["fields"]:
                    continue
                lines.append(
                    "        <td style=\"padding-left: %dem; padding-right: %dem; text-align: %s; vertical-align: %s\">%s</td>" % (
                        lpad, rpad, align, valign, escape(datum).replace("\n", linebreak)))
            lines.append("    </tr>")
        lines.append("</table>")

        return self._unicode("\n").join(lines)


##############################
# UNICODE WIDTH FUNCTIONS    #
##############################

def _char_block_width(char):
    """

    :param char:

    """
    # Basic Latin, which is probably the most common case
    # if char in xrange(0x0021, 0x007e):
    # if char >= 0x0021 and char <= 0x007e:
    if 0x0021 <= char <= 0x007e:
        return 1
    # Chinese, Japanese, Korean (common)
    if 0x4e00 <= char <= 0x9fff:
        return 2
    # Hangul
    if 0xac00 <= char <= 0xd7af:
        return 2
    # Combining?
    if unicodedata.combining(uni_chr(char)):
        return 0
    # Hiragana and Katakana
    if 0x3040 <= char <= 0x309f or 0x30a0 <= char <= 0x30ff:
        return 2
    # Full-width Latin characters
    if 0xff01 <= char <= 0xff60:
        return 2
    # CJK punctuation
    if 0x3000 <= char <= 0x303e:
        return 2
    # Backspace and delete
    if char in (0x0008, 0x007f):
        return -1
    # Other control characters
    elif char in (0x0000, 0x000f, 0x001f):
        return 0
    # Take a guess
    return 1


def _str_block_width(val):
    """

    :param val:

    """
    return sum(itermap(_char_block_width, itermap(ord, _re.sub("", val))))



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Unsupervised machine learning algorithm
=======================================

The :mod:`~gemseo.mlearning.core.unsupervised` module implements the concept of
unsupervised machine learning models, where the data has no notion of
input or output.

This concept is implemented through the :class:`.MLUnsupervisedAlgo` class,
which inherits from the :class:`.MLAlgo` class.
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library
from numpy import hstack

from gemseo.mlearning.core.ml_algo import MLAlgo

standard_library.install_aliases()


class MLUnsupervisedAlgo(MLAlgo):
    """Unsupervised machine learning algorithm.

    Inheriting classes should overload the
    :meth:`!MLUnsupervisedAlgo._fit` method.
    """

    ABBR = "MLUnupervisedAlgo"

    def __init__(self, data, transformer=None, var_names=None, **parameters):
        """Constructor.

        :param Dataset data: learning dataset
        :param transformer: transformation strategy for data groups.
            If None, do not scale data. Default: None.
        :type transformer: dict(str)
        :param var_names: names of the variables to consider.
        :type var_names: list(str)
        :param parameters: algorithm parameters
        """
        super(MLUnsupervisedAlgo, self).__init__(
            data, transformer=transformer, var_names=var_names, **parameters
        )
        self.var_names = var_names or data.variables

    def learn(self, samples=None):
        """Train machine learning algorithm on learning set.

        :param list(str) names: learning variables. Default: None.
        :param list(int) samples: training samples (indices). Default: None.
        """
        if set(self.var_names) == set(self.learning_set.variables):
            data = []
            for group in self.learning_set.groups:
                sub_data = self.learning_set.get_data_by_group(group)
                if group in self.transformer:
                    sub_data = self.transformer[group].fit_transform(sub_data)
                data.append(sub_data)
            data = hstack(data)
        else:
            data = []
            for name in self.var_names:
                sub_data = self.learning_set.get_data_by_names([name], False)
                if name in self.transformer:
                    sub_data = self.transformer[name].fit_transform(sub_data)
                data.append(sub_data)
            data = hstack(data)

        if samples is not None:
            data = data[samples]

        self._fit(data)
        self._trained = True

    def _fit(self, data):
        """Fit model on data.

        :param ndarray data: training data (2D).
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Matthias De Lozzo, Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Jameson sensor
==============
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library
from numpy import abs as np_abs
from numpy import amax

from gemseo.mlearning.transform.transformer import Transformer

standard_library.install_aliases()


class JamesonSensor(Transformer):
    """ Jameson Sensor. """

    def __init__(
        self, name="JamesonSensor", threshold=0.3, removing_part=0.01, dimension=1
    ):
        """Constructor.

        :param str name: name of the sensor. Default: 'JamesonSensor'.
        :param float threshold: value to add to the denominator
            to avoid zero division. Default: 0.3.
        :param float removing_part: define the level of the signal to
            remove in order to avoid leading and trailing edge effects.
            ONLY FOR 1D MESH, to redefine for 2D mesh. Default: 0.01.
        :param int dimension: mesh dimension. Default: 1.
        """
        super(JamesonSensor, self).__init__(name)
        self.threshold = threshold
        self.removing_part = removing_part
        self.dimension = dimension

    def fit(self, data):
        """Fit sensor to data.

        :param array data: data to be fitted.
        """
        self.threshold = self.threshold * amax(data)

    def transform(self, data):
        """Transform data.

        :param ndarray data: data  to be transformed.
        :return: transformed data.
        :rtype: ndarray
        """
        if self.dimension == 1:
            transformed_data = self._jameson_1_d(data)
        else:
            raise NotImplementedError()

        return transformed_data

    def _jameson_1_d(self, data):
        """Transform data.

        :param ndarray data: data  to be transformed.
        :return: transformed data.
        :rtype: ndarray
        """
        mesh_size = data.shape[1] - 2
        min_mesh_size = int(mesh_size * self.removing_part)
        max_mesh_size = int(mesh_size * (1 - self.removing_part))
        norm = np_abs(data[:, :-2])
        norm += 2 * np_abs(data[:, 1:-1])
        norm += np_abs(data[:, 2:])
        norm = norm + self.threshold
        result = abs(data[:, :-2] - 2 * data[:, 1:-1] + data[:, 2:]) / norm
        result = result[:, min_mesh_size:max_mesh_size]
        return result



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or
#                      initial documentation
#        :author:  Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Factory base class
******************
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import importlib
import os
import pkgutil
import sys

from future import standard_library
from future.utils import with_metaclass

from gemseo.core.json_grammar import JSONGrammar
from gemseo.third_party.prettytable import PrettyTable
from gemseo.utils.singleton import SingleInstancePerAttributeEq
from gemseo.utils.source_parsing import SourceParsing

standard_library.install_aliases()

from gemseo import LOGGER


class Factory(with_metaclass(SingleInstancePerAttributeEq, object)):
    """
    Factory to create extensions that are known to |g|:
        can be a MDODiscipline, MDOFormulation... Depending on the subclass

    Three types of directories are scanned :

    - the environment variable "GEMSEO_PATH" may contain
      the list of directories to scan
    - internal_modules_paths (such as gemseo.problems...)
    - a directory list may be passed to the factory
    """

    # Name of the environment variable to search for classes
    GEMSEO_PATH = "GEMSEO_PATH"
    GEMS_PATH = "GEMS_PATH"

    # Allowed prefix for naming a plugin in sys.path
    PLUGIN_PREFIX = "gemseo_"

    def __init__(self, base_class=None, internal_modules_paths=None):
        """Initializes the factory.

        Scans the directories to search for subclasses of MDODiscipline.
        Searches in "GEMSEO_PATH", "GEMS_PATH",  and gemseo.problems

        :param base_class: class to search in the modules
            (MDOFormulation, MDODiscipline...) depending on the subclass
        :param internal_modules_paths: import paths (such as gemseo.problems)
            which are already imported
        :param name: name of the factory to print when configuration
                     is printed
        :param possible_plugin_names: tuple of plugins packages names to be
            scanned if they can be imported. The last plugin name has the
            priority. For instance, if the same class MDAJacobi exists in
            gemseo.mda, gemseo_plugins.mda and gemseo_private.mda, the used one will
            be gemseo_private.mda
        """
        if not isinstance(base_class, type):
            raise TypeError("Class to search must be a class!")

        self.base_class = base_class
        self.internal_modules_paths = internal_modules_paths or []

        self.failed_imports = {}
        self.__names_to_classes = {}

        self.update()

    def _update_path_from_env_variable(self, env_variable):
        """Update the classes that can be instanciated from a factory from an
        environment variable.

        param env_variable: name of the environment variable
        """
        g_path = os.environ.get(env_variable)
        if g_path is None:
            return

        if ":" in g_path:
            paths = g_path.split(":")
        else:
            paths = [g_path]

        # temporary make the gemseo paths visible to the import machinery
        for path in paths:
            sys.path.insert(0, path)
        for _, mod_name, _ in pkgutil.iter_modules(path=paths):
            self.__import_modules_from(mod_name)
        for path in paths:
            sys.path.pop(0)

    def update(self):
        """Update the classes that can be created by the factory.

        In order, scan in the internal modules, then in plugins, then in
        GEMSEO_PATH.
        """
        # Scan internal packages
        for mod_name in self.internal_modules_paths:
            self.__import_modules_from(mod_name)

        # Scan plugins packages
        for _, mod_name, _ in pkgutil.iter_modules():
            if mod_name.startswith(self.PLUGIN_PREFIX):
                self.__import_modules_from(mod_name)

        gems_path = os.environ.get(self.GEMS_PATH)
        if gems_path is not None:
            msg = """GEMS is now named GEMSEO. The GEMS_PATH environment
             variable is now deprecated and it is strongly recommended to use
             the GEMSEO_PATH environment variable instead to register your
             GEMSEO plugins."""
            LOGGER.warn(msg)

        # Scan environment variable paths
        env_variables = [self.GEMSEO_PATH, self.GEMS_PATH]
        for env_variable in env_variables:
            self._update_path_from_env_variable(env_variable)

        self.__names_to_classes = self.__get_sub_classes(self.base_class)

    def __log_import_failure(self, pkg_name):
        """Log import failures."""
        LOGGER.debug("Failed to import package %s", pkg_name)
        self.failed_imports[pkg_name] = ""

    def __import_modules_from(self, pkg_name):
        """Import modules from a package."""
        pkg = importlib.import_module(pkg_name)

        if not hasattr(pkg, "__path__"):
            # not a package so no more module to import
            return

        for _, mod_name, _ in pkgutil.walk_packages(
            pkg.__path__, pkg.__name__ + ".", self.__log_import_failure
        ):
            try:
                importlib.import_module(mod_name)
            except Exception as err:  # pylint: disable=(broad-except
                LOGGER.debug("Failed to import module: %s, %s", mod_name, err)
                self.failed_imports[mod_name] = err

    def __get_sub_classes(self, cls):
        """Return all the sub classes of cls.

        The class names are unique, the last imported is kept when more than
        one class have the same name.
        """
        all_sub_classes = {}
        for sub_class in cls.__subclasses__():
            sub_classes = {sub_class.__name__: sub_class}
            sub_classes.update(self.__get_sub_classes(sub_class))
            for cls_name, _cls in sub_classes.items():
                all_sub_classes[cls_name] = _cls
        return all_sub_classes

    @property
    def classes(self):
        """Return the available classes.

        :returns : the list of classes names
        """
        return sorted(self.__names_to_classes.keys())

    # TODO: rename to has_class
    def is_available(self, name):
        """Return whether a class is available.

        :param name : name of the class
        :returns: True if the class is available
        """
        return name in self.__names_to_classes

    def get_class(self, name):
        """Return the class from its name.

        :param name : name of the class
        """
        try:
            return self.__names_to_classes[name]
        except KeyError:
            msg = "Class {} is not available!\nAvailable ones are: {}".format(
                name, ", ".join(sorted(self.__names_to_classes.keys()))
            )
            raise ImportError(msg)

    def create(self, class_name, **options):
        """Return an instance with given class name.

        :param class_name : name of the class
        :parma options: options to be passed to the constructor
        """
        cls = self.get_class(class_name)
        try:
            return cls(**options)
        except TypeError:
            # TODO: raise an error with message and let the callers handle
            # logging
            LOGGER.error(
                "Failed to create class %s with arguments %s", class_name, options
            )
            raise

    def get_options_doc(self, name):
        """Return the options documentation for the given class name.

        :param name: name of the class
        :returns: the dict of option name: option documentation
        """
        cls = self.get_class(name)
        return SourceParsing.get_options_doc(cls.__init__)

    def get_default_options_values(self, name):
        """Return the options default values for the given class name.

        Only addresses kwargs

        :param name : name of the class
        :returns: the dict option name: option default value
        """
        cls = self.get_class(name)
        return SourceParsing.get_default_options_values(cls)

    def get_options_grammar(self, name, write_schema=False, schema_file=None):
        """Return the options grammar for a class.

        Attempts to generate a JSONGrammar from the arguments of the __init__
        method of the class

        :param name: name of the class
        :param schema_file: the output json file path. If None: input.json or
            output.json depending on gramamr type.
            (Default value = None)
        :param write_schema: if True, writes the schema files
            (Default value = False)
        :returns: the json grammar for options
        """
        args_dict = self.get_default_options_values(name)
        opts_doc = self.get_options_doc(name)
        opts_doc = {k: v for k, v in opts_doc.items() if k in args_dict}
        gramm = JSONGrammar(name)

        gramm.initialize_from_base_dict(
            args_dict,
            schema_file=schema_file,
            write_schema=write_schema,
            description_dict=opts_doc,
        )
        # Remove None args from required
        sch_dict = gramm.schema.to_dict()
        required = sch_dict["required"]
        has_changed = False
        for opt, val in args_dict.items():
            if val is None and opt in required:
                required.remove(opt)
                has_changed = True
        if has_changed:
            gramm = JSONGrammar(name, schema=sch_dict)
        return gramm

    def get_sub_options_grammar(self, class_name, **options):
        """Return the JSONGrammar of the sub options of a class.

        :param class_name: name of the class
        :param options: options to be passed to the class required to deduce
            the sub options
        """
        cls = self.get_class(class_name)
        return cls.get_sub_options_grammar(**options)

    def get_default_sub_options_values(self, class_name, **options):
        """Return the default values of the sub options of a class.

        :param class_name: name of the class
        :param options: options to be passed to the class required to deduce
            the sub options
        """
        cls = self.get_class(class_name)
        return cls.get_default_sub_options_values(**options)

    def __str__(self):
        """Return the representation of a factory.

        Gives the configuration with the successfully loaded modules and
        failed imports with the reason.
        """
        table = PrettyTable(
            ["Module", "Is available ?", "Purpose or error message"],
            title=self.base_class.__name__,
            min_table_width=120,
            max_table_width=120,
        )

        row_dict = {}
        for cls in self.__names_to_classes.values():
            msg = ""
            try:
                msgs = cls.__doc__.split("\n")
                while msgs and msg == "":
                    msg = msgs[0]
                    del msgs[0]
            except Exception as err:  # pylint: disable=(broad-except
                pass

            key = cls.__name__
            row_dict[key] = [key, "Yes", msg]

        for key, err in self.failed_imports.items():
            row_dict[key] = [key, "No", str(err)]

        # Take them all and then sort them for pretty printing
        for key in sorted(row_dict.keys()):
            table.add_row(row_dict[key])

        return table.get_string()



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Clustering algorithm
====================

The :mod:`~gemseo.mlearning.cluster.cluster` module
implements the concept of clustering models,
a kind of unsupervised machine learning algorithm where the goal is
to group data into clusters.
Wherever it is possible, these methods should be able to predict the class of
new data, as well as the probability of belonging to each class.

This concept is implemented through the :class:`.MLClusteringAlgo` class
which inherits from the :class:`.MLUnsupervisedAlgo` class.
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library
from numpy import atleast_2d, unique, zeros

from gemseo.mlearning.core.unsupervised import MLUnsupervisedAlgo
from gemseo.utils.data_conversion import DataConversion

standard_library.install_aliases()


class MLClusteringAlgo(MLUnsupervisedAlgo):
    """Clustering algorithm.

    Inheriting class should overload the
    :meth:`!MLUnsupervisedAlgo._fit` method, and the
    :meth:`!MLClusteringAlgo._predict` and
    :meth:`!MLClusteringAlgo._predict_proba` methods if possible.
    """

    def __init__(self, data, transformer=None, var_names=None, **parameters):
        """Constructor.

        :param Dataset data: learning dataset.
        :param transformer: transformation strategy for data groups.
            If None, do not scale data. Default: None.
        :type transformer: dict(str)
        :param var_names: names of the variables to consider.
        :type var_names: list(str)
        :param parameters: algorithm parameters.
        """
        super(MLClusteringAlgo, self).__init__(
            data, transformer=transformer, var_names=var_names, **parameters
        )
        self.labels = None
        self.n_clusters = None

    def learn(self, samples=None):
        """Overriding learn function for assuring that labels are defined.
        Identify number of clusters.
        """
        super(MLClusteringAlgo, self).learn(samples=samples)
        if self.labels is None:
            raise ValueError("self._fit() should assign labels.")
        self.n_clusters = unique(self.labels).shape[0]

    def predict(self, data):
        """Predict cluster of data.

        :param data: data (1D or 2D).
        :type data: dict(ndarray) or ndarray
        :return: clusters of data ("0D" or 1D).
        :rtype: int or ndarray(int)
        """
        as_dict = isinstance(data, dict)
        if as_dict:
            data = DataConversion.dict_to_array(data, self.var_names)
        single_sample = len(data.shape) == 1
        data = atleast_2d(data)
        parameters = self.learning_set.DEFAULT_GROUP
        if parameters in self.transformer:
            data = self.transformer[parameters].transform(data)
        clusters = self._predict(atleast_2d(data)).astype(int)
        if single_sample:
            clusters = clusters[0]
        return clusters

    def _predict(self, data):
        """Predict cluster of data.

        :param ndarray data: data (2D).
        :return: clusters of data (1D).
        :rtype: ndarray(int)
        """
        raise NotImplementedError

    def predict_proba(self, data, hard=True):
        """Predict probability of belonging to each cluster.

        :param data: data (1D or 2D).
        :type data: dict(ndarray) or ndarray
        :param bool hard: indicator for hard or soft clustering. Default: True.
        :return: probabilities of belonging to each cluster (1D or 2D, same
            as data).
        :rtype: ndarray
        """
        as_dict = isinstance(data, dict)
        if as_dict:
            data = DataConversion.dict_to_array(data, self.var_names)
        single_sample = len(data.shape) == 1
        data = atleast_2d(data)
        probas = self._predict_proba(atleast_2d(data), hard)
        if single_sample:
            probas = probas.ravel()
        return probas

    def _predict_proba(self, data, hard=True):
        """Predict probability of belonging to each cluster.

        :param ndarray data: data (2D).
        :param bool hard: indicator for hard or soft clustering. Default: True.
        :return: probabilities of belonging to each cluster (2D). The sum of
            each row is one.
        :rtype: ndarray
        """
        if hard:
            probas = self._predict_proba_hard(data)
        else:
            probas = self._predict_proba_soft(data)
        return probas

    def _predict_proba_hard(self, data):
        """Create cluster indicator of input data.

        :param ndarray input_data: input data (2D).
        :return: cluster indicators for each sample (2D). The sum of
            each row is one.
        :rtype: ndarray
        """
        prediction = self._predict(data)
        probas = zeros((data.shape[0], self.n_clusters))
        for i, pred in enumerate(prediction):
            probas[i, pred] = 1
        return probas

    def _predict_proba_soft(self, data):
        """Predict probability of belonging to each cluster.

        :param ndarray data: data (2D).
        :return: probabilities for each cluster for each sample (2D). The sum
            of each row is one.
        :rtype: ndarray
        """
        raise NotImplementedError

    def _get_objects_to_save(self):
        """Get objects to save.
        :return: objects to save.
        :rtype: dict
        """
        objects = super(MLClusteringAlgo, self)._get_objects_to_save()
        objects["labels"] = self.labels
        objects["n_clusters"] = self.n_clusters
        return objects



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Mean squared error measure
==========================

The :mod:`~gemseo.mlearning.qual_measure.mse_measure` module
implements the concept of means squared error measures
for machine learning algorithms.

This concept is implemented through the
:class:`.MSEMeasure` class and
overloads the :meth:`!MLErrorMeasure._compute_measure` method.

The mean squared error (MSE) is defined by

.. math::

    \\operatorname{MSE}(\\hat{y})=\\frac{1}{n}\\sum_{i=1}^n(\\hat{y}_i-y_i)^2,

where
:math:`\\hat{y}` are the predictions and
:math:`y` are the data points.
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library
from sklearn.metrics import mean_squared_error

from gemseo.mlearning.qual_measure.error_measure import MLErrorMeasure

standard_library.install_aliases()


class MSEMeasure(MLErrorMeasure):
    """ Mean Squared Error measure for machine learning. """

    def _compute_measure(self, outputs, predictions, multioutput=True):
        """Compute MSE.

        :param ndarray outputs: reference outputs.
        :param ndarray predictions: predicted outputs.
        :param bool multioutput: if True, return the error measure for each
            output component. Otherwise, average these errors. Default: True.
        :return: MSE value.
        """
        multioutput = "raw_values" if multioutput else "uniform_average"
        return mean_squared_error(outputs, predictions, multioutput=multioutput)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Baseclass for all formulations
******************************
"""

from __future__ import absolute_import, division, print_function, unicode_literals

from future import standard_library
from numpy import array, copy, empty, in1d, where, zeros
from six import string_types

from gemseo.algos.opt_problem import OptimizationProblem
from gemseo.core.function import MDOFunction, MDOFunctionGenerator

standard_library.install_aliases()


from gemseo import LOGGER


class MDOFormulation(object):
    """Abstract MDO formulation class
    To be extended in subclasses for use.

    The MDOFormulation creates the objective function and
    constraints from the disciplines.

    It defines the process implicitly.

    The link between MDODisciplines and objective
    functions and constraints is made with MDOFunctionGenerator,
    which generates MDOFunctions from the disciplines.
    """

    NAME = "MDOFormulation"

    def __init__(
        self,
        disciplines,
        objective_name,
        design_space,
        maximize_objective=False,
        **options
    ):  # pylint: disable=W0613
        """
        Constructor, initializes the objective functions and constraints

        :param disciplines: the disciplines list
        :param reference_input_data: the base input data dict for the
            disciplines
        :param objective_name: the objective function data name
        :param design_space: the design space
        :param maximize_objective: if True, the objective function
            is maximized, by default, a minimization is performed
        """
        self.check_disciplines(disciplines)
        self.disciplines = disciplines
        self._objective_name = objective_name
        self.opt_problem = OptimizationProblem(design_space)
        self._maximize_objective = maximize_objective

    @property
    def design_space(self):
        """
        Proxy for formulation.design_space

        :returns: the design space
        """
        return self.opt_problem.design_space

    @staticmethod
    def check_disciplines(disciplines):
        """Sets the disciplines.

        :param disciplines: the disciplines list
        """
        if not disciplines or not isinstance(disciplines, list):
            raise TypeError(
                "Disciplines must be provided" + " to the formulation as a list"
            )

    @staticmethod
    def _check_add_cstr_input(output_name, constraint_type):
        """Checks the add_constraint method inputs
        Can be reused in subclasses

        :param output_name: the output name to be used as constraint
            for instance, if g_1 is given and constraint_type="eq",
            g_1=0 will be added as constraint to the optimizer
        :param constraint_type: the type of constraint, "eq" for equality,
            "ineq" for inequality constraint
        """
        if constraint_type not in [MDOFunction.TYPE_EQ, MDOFunction.TYPE_INEQ]:
            raise ValueError(
                "Constraint type must be either 'eq' or 'ineq',"
                + " got:"
                + str(constraint_type)
                + " instead"
            )
        if isinstance(output_name, list):
            outputs_list = output_name
        else:
            outputs_list = [output_name]
        return outputs_list

    def add_constraint(
        self,
        output_name,
        constraint_type=MDOFunction.TYPE_EQ,
        constraint_name=None,
        value=None,
        positive=False,
    ):
        """Add a user constraint, i.e. a design constraint in addition to
        formulation specific constraints such as targets in IDF.
        The strategy of repartition of constraints is defined in the
        formulation class.

        :param output_name: the output name to be used as constraint
            for instance, if g_1 is given and constraint_type="eq",
            g_1=0 will be added as constraint to the optimizer
        :param constraint_type: the type of constraint, "eq" for equality,
            "ineq" for inequality constraint
            (Default value = MDOFunction.TYPE_EQ)
        :param constraint_name: name of the constraint to be stored,
            if None, generated from the output name (Default value = None)
        :param value: Default value = None)
        :param positive: Default value = False)
        :returns: the constraint as an MDOFunction
            False if the formulation does not dispatch the constraint to the
            optimizers itself
        """
        outputs_list = self._check_add_cstr_input(output_name, constraint_type)

        mapped_cstr = self._get_function_from(outputs_list, top_level_disc=True)
        mapped_cstr.f_type = constraint_type

        if constraint_name is not None:
            mapped_cstr.name = constraint_name
        self.opt_problem.add_constraint(mapped_cstr, value=value, positive=positive)

    def add_observable(self, output_names, observable_name=None, discipline=None):
        """
        Adds observable to the optimization problem. The repartition
        strategy of the observable is defined in the formulation class.

        :param output_names: names of the outputs to observe
        :type output_names: str or list(str)
        :param observable_name: name of the observable, optional.
        :type observable_name: str
        :param discipline: if None, detected from inner disciplines, otherwise
            the discipline used to build the function
            (Default value = None)
        :type discipline: MDODiscipline
        """
        if isinstance(output_names, string_types):
            output_names = [output_names]
        obs_fun = self._get_function_from(
            output_names, top_level_disc=True, discipline=discipline
        )
        if observable_name is not None:
            obs_fun.name = observable_name
        obs_fun.f_type = MDOFunction.TYPE_OBS
        self.opt_problem.add_observable(obs_fun)

    def get_top_level_disc(self):
        """Returns the disciplines which inputs are required to run the
        associated scenario
        By default, returns all disciplines
        To be overloaded by subclasses

        :returns: the list of top level disciplines
        """
        return self.disciplines

    @staticmethod
    def _get_mask_from_datanames(all_data_names, masked_data_names):
        """Gets a mask of all_data_names for masked_data_names, ie an array
        of the size of all_data_names with True values when masked_data_names
        are in all_data_names

        :param all_data_names: the main array for mask
        :param masked_data_names: the array which masks all_data_names
        :returns: a True / False valued mask array
        """
        places = in1d(all_data_names, masked_data_names)
        return where(places)

    def _get_generator_from(self, output_names, top_level_disc=False):
        """Find a discipline which has all outputs named output_names
        and builds the associated MDOFunctionGenerator

        :param top_level_disc: if True, search outputs among top
            level disciplines (Default value = False)
        :param output_names: the output names
        :returns: the discipline
        """
        if top_level_disc:
            search_among = self.get_top_level_disc()
        else:
            search_among = self.disciplines
        for discipline in search_among:
            if discipline.is_all_outputs_existing(output_names):
                return MDOFunctionGenerator(discipline)
        raise ValueError(
            "No discipline known by formulation "
            + type(self).__name__
            + " has all outputs named "
            + str(output_names)
        )

    def _get_generator_with_inputs(self, input_names, top_level_disc=False):
        """Find a discipline which has all inputs named output_names

        :param input_names: the output names
        :param top_level_disc: if True, search outputs among top
            level disciplines (Default value = False)
        :returns: the discipline
        """
        if top_level_disc:
            search_among = self.get_top_level_disc()
        else:
            search_among = self.disciplines
        for discipline in search_among:
            if discipline.is_all_inputs_existing(input_names):
                return MDOFunctionGenerator(discipline)
        raise ValueError(
            "No discipline known by formulation "
            + type(self).__name__
            + " has all inputs named "
            + str(input_names)
        )

    def mask_x(self, masking_data_names, x_vect, all_data_names=None):
        """Masks a vector x_vect, using names masking_data_names,
        and with respect to reference names all_data_names

        :param masking_data_names: the names of data to keep
        :param x_vect: the vector to mask
        :param all_data_names: reference data names, if None,
            self.get_optim_variables_names() used instead
            (Default value = None)
        :returns: masked x_vect
        """
        if all_data_names is None:
            all_data_names = self.get_optim_variables_names()
        i_min = 0
        x_mask = array([False] * x_vect.size)
        for key in all_data_names:
            var_length = self._get_dv_length(key)
            i_max = i_min + var_length
            if len(x_vect) < i_max:
                msg = "Inconsistent input size array " + str(key) + " = "
                msg += str(x_vect.shape) + " for the design variable "
                msg += "of length " + str(var_length)
                raise ValueError(msg)
            if key in masking_data_names:
                x_mask[i_min:i_max] = True
            i_min = i_max

        return x_vect[x_mask]

    def unmask_x(self, masking_data_names, x_masked, all_data_names=None, x_full=None):
        """Unmasks a vector x, using names masking_data_names,
        and with respect to
        reference names all_data_names

        :param masking_data_names: the names of data to keep
        :param x_masked: the vector to unmask
        :param all_data_names: reference data names, if None,
            self.get_optim_variables_names() used instead
            (Default value = None)
        :param x_full: the default values for the full vector, if None,
            np.zeros() is used
        :returns: unmasked x
        """
        return self.unmask_x_swap_order(
            masking_data_names, x_masked, all_data_names, x_full
        )

    def _get_dv_length(self, variable_name):
        """Retrieves the length of a variable from the size declared
        in the design space

        :param variable_name: name of the variable

        """
        return self.opt_problem.design_space.get_size(variable_name)

    def _get_x_mask_swap(self, masking_data_names, all_data_names=None):
        """gets the mask dict, using names masking_data_names,
        and with respect to reference names all_data_names
        eventually swaps the order of the values if data names order
        are inconsistent between masking_data_names and all_data_names

        :param masking_data_names: the names of data to keep
        :param all_data_names: reference data names, if None,
            self.get_optim_variables_names() used instead
            (Default value = None)
        :returns: mask dict
        """
        if all_data_names is None:
            all_data_names = self.get_optim_variables_names()
        i_min = 0
        x_values_dict = {}
        n_x = 0
        for key in all_data_names:
            i_max = i_min + self._get_dv_length(key)
            if key in masking_data_names:
                x_values_dict[key] = (i_min, i_max)
                n_x += i_max - i_min
            i_min = i_max
        return x_values_dict, n_x, i_max

    def unmask_x_swap_order(
        self, masking_data_names, x_masked, all_data_names=None, x_full=None
    ):
        """Unmasks a vector x, using names masking_data_names,
        and with respect to  reference names all_data_names
        eventually swaps the order of the x values if data names order
        are inconsistent between masking_data_names and all_data_names

        :param masking_data_names: the names of data to keep
        :param x_masked: the masked vector
        :param all_data_names: reference data names, if None,
            self.get_optim_variables_names() used instead
            (Default value = None)
        :param x_full: the default values for the full vector, if None,
                        np.zeros() is used
        :returns: unmasked x
        """
        if all_data_names is None:
            all_data_names = self.get_optim_variables_names()
        x_values_dict, _, len_x = self._get_x_mask_swap(
            masking_data_names, all_data_names
        )
        if x_full is None:
            x_unmask = zeros((len_x), dtype=x_masked.dtype)
        else:
            x_unmask = copy(x_full)

        i_x = 0
        for key in all_data_names:
            if key in x_values_dict:
                i_min, i_max = x_values_dict[key]
                n_x = i_max - i_min
                if x_masked.size < i_x + n_x:
                    msg = "Inconsistent data shapes !\nTry to unmask data "
                    msg += key + " of len " + str(n_x)
                    msg += "\nWith values of len : " + str(x_masked.size)
                    raise ValueError(msg)
                x_unmask[i_min:i_max] = x_masked[i_x : i_x + n_x]
                i_x += n_x
        return x_unmask

    def mask_x_swap_order(self, masking_data_names, x_vect, all_data_names=None):
        """Masks a vector x_vect, using names masking_data_names,
        and with respect to reference names all_data_names
        possibly swaps the order of the x_vect values if data names orders
        are inconsistent between masking_data_names and all_data_names

        :param masking_data_names: the names of data to keep
        :param x_vect: the vector to mask
        :param all_data_names: reference data names, if None,
            self.get_optim_variables_names() used instead
            (Default value = None)
        :returns: masked x_vect
        """
        if all_data_names is None:
            all_data_names = self.get_optim_variables_names()
        x_values_dict, n_x, _ = self._get_x_mask_swap(
            masking_data_names, all_data_names
        )
        x_masked = zeros((n_x), dtype=x_vect.dtype)
        i_max = 0
        i_min = 0
        for key in masking_data_names:
            if key not in x_values_dict:
                raise ValueError(
                    "Inconsistent inputs of masking. Key "
                    + str(key)
                    + " is in masking_data_names"
                    + str(masking_data_names)
                    + " but not in provided all_data_names :"
                    + str(all_data_names)
                    + " !"
                )
            value = x_values_dict[key]
            i_max += value[1] - value[0]
            len_x = len(x_vect)
            if len(x_masked) < i_max or len_x <= value[0] or len_x < value[1]:
                raise ValueError(
                    "Inconsistent input array size of values "
                    + "array "
                    + str(x_vect)
                    + " with reference data shape "
                    + str(x_vect.shape)
                    + ", for data named :"
                    + str(key)
                    + " of size : "
                    + str(i_max)
                )
            x_masked[i_min:i_max] = x_vect[value[0] : value[1]]
            i_min = i_max

        return x_masked

    def _remove_unused_variables(self):
        """Removes variables in the design space that are not discipline inputs"""
        design_space = self.opt_problem.design_space
        disciplines = self.get_top_level_disc()
        all_inputs = set(
            var for disc in disciplines for var in disc.get_input_data_names()
        )
        for name in set(design_space.variables_names):
            if name not in all_inputs:
                design_space.remove_variable(name)

    def _remove_sub_scenario_dv_from_ds(self):
        """Removes the sub scenarios design variables from the design space"""
        for scenario in self.get_sub_scenarios():
            loc_vars = scenario.design_space.variables_names
            for var in loc_vars:
                if var in self.design_space.variables_names:
                    self.design_space.remove_variable(var)

    def _build_objective_from_disc(
        self, objective_name, discipline=None, top_level_disc=True
    ):
        """Given the objective name, finds the discipline which is able to
        compute it and builds the objective function from it.

        :param objective_name: the name of the objective
        :param discipline: if None, detected from inner disciplines, otherwise
            the discipline used to build the function
            (Default value = None)
        :param top_level_disc: Default value = True)
        """
        if isinstance(objective_name, string_types):
            objective_name = [objective_name]
        obj_mdo_fun = self._get_function_from(
            objective_name, discipline, top_level_disc=top_level_disc
        )
        obj_mdo_fun.f_type = MDOFunction.TYPE_OBJ
        self.opt_problem.objective = obj_mdo_fun
        if self._maximize_objective:
            self.opt_problem.change_objective_sign()

    def _get_function_from(
        self,
        output_names,
        discipline=None,
        top_level_disc=True,
        x_names=None,
        all_data_names=None,
    ):
        """Builds the objective function from a discipline.

        :param output_names: the name list of the  data
        :param discipline: if None, detected from inner disciplines, otherwise
            the discipline used to build the function
            (Default value = None)
        :param top_level_disc: if True, only high level disciplines
            are used to select discipline to build
            the function (Default value = True)
        :param x_names: names of the design variables, if None, use
            self.get_x_names_of_disc(discipline)
            (Default value = None)
        :param all_data_names: reference data names for masking x, if None,
            self.get_optim_variables_names() used instead
            (Default value = None)
        """
        if discipline is None:
            gen = self._get_generator_from(output_names, top_level_disc=top_level_disc)
            discipline = gen.discipline
        else:
            gen = MDOFunctionGenerator(discipline)

        if x_names is None:
            x_names = self.get_x_names_of_disc(discipline)

        out_x_func = gen.get_function(x_names, output_names)

        def func(x_vect):
            """Function to compute consistency constraints

            :param x: design variable vector
            :param x_vect: returns: value of consistency constraints
                    (=0 if disciplines are at equilibrium)
            :returns: value of consistency constraints
                    (=0 if disciplines are at equilibrium)
            """
            x_of_disc = self.mask_x_swap_order(x_names, x_vect, all_data_names)
            obj_allx_val = out_x_func(x_of_disc)
            return obj_allx_val

        masked_func = MDOFunction(
            func,
            out_x_func.name,
            f_type=MDOFunction.TYPE_OBJ,
            args=x_names,
            expr=out_x_func.expr,
            dim=out_x_func.dim,
            outvars=out_x_func.outvars,
        )
        if out_x_func.has_jac():

            def func_jac(x_vect):
                """Function to compute consistency constraints gradient

                :param x: design variable vector
                :param x_vect: returns: gradient of consistency constraints
                :returns: gradient of consistency constraints
                """
                x_of_disc = self.mask_x_swap_order(x_names, x_vect, all_data_names)

                loc_jac = out_x_func.jac(x_of_disc)  # pylint: disable=E1102

                if len(loc_jac.shape) == 1:
                    # This is surprising but there is a duality between the
                    # masking operation in the function inputs and the
                    # unmasking of its outputs
                    jac = self.unmask_x_swap_order(x_names, loc_jac, all_data_names)
                else:
                    n_outs = loc_jac.shape[0]
                    jac = empty((n_outs, x_vect.size), dtype=x_vect.dtype)
                    for func_ind in range(n_outs):
                        gr_u = self.unmask_x_swap_order(
                            x_names, loc_jac[func_ind, :], all_data_names
                        )
                        jac[func_ind, :] = gr_u
                return jac

            masked_func.jac = func_jac

        return masked_func

    def get_optim_variables_names(self):
        """Gets the optimization unknown names to be provided to the optimizer
        This is different from the design variable names provided by the user,
        since it depends on the formulation, and can include target values
        for coupling for instance in IDF

        :returns: optimization unknown names
        """
        return self.opt_problem.design_space.variables_names

    def get_x_names_of_disc(self, discipline):
        """Gets the design variables names of a given discipline

        :param discipline: the discipline
        :returns: design variables names
        """
        optim_variables_names = self.get_optim_variables_names()
        input_names = discipline.get_input_data_names()
        return [name for name in optim_variables_names if name in input_names]

    def get_sub_disciplines(self):
        """Accessor to the sub disciplines.
        Lists the sub scenarios' disciplines.

        :returns: list of sub disciplines
        """
        sub_disc = []

        def add_to_sub(disc_list):
            """
            Adds the discipline list to the sub_sc disc list
            if not already in it

            :param disc_list: the discipline list
            """
            for disc in disc_list:
                if disc not in sub_disc:
                    sub_disc.append(disc)

        for discipline in self.disciplines:
            if hasattr(discipline, "disciplines"):
                add_to_sub(discipline.disciplines)
            else:
                add_to_sub([discipline])

        return sub_disc

    def get_sub_scenarios(self):
        """Lists the disciplines that are actually scenarios

        :returns: the list of scenarios
        """
        return [disc for disc in self.disciplines if disc.is_scenario()]

    def _set_defaultinputs_from_ds(self):
        """
        Sets the default inputs of the top level disciplines
        from the design space
        """
        if not self.opt_problem.design_space.has_current_x():
            return
        x_dict = self.opt_problem.design_space.get_current_x_dict()
        for disc in self.get_top_level_disc():
            inputs = disc.get_input_data_names()
            curr_x_disc = {name: x_dict[name] for name in inputs if name in x_dict}
            disc.default_inputs.update(curr_x_disc)

    def get_expected_workflow(self):
        """Returns the sequence of execution of the disciplines to be expected
        regarding the implementation.
        The returned value is an array containing disciplines,
        tuples of disciplines for concurrent execution.
        For instance :
        * [A, B] denotes the execution of A then the execution of B
        * (A, B) denotes the concurrent execution of A and B
        * [A, (B, C), D] denotes the execution of A then the concurrent
        execution of B and C then the execution of D.

        :returns: array containing disciplines or tuples of disciplines.
            Used for xdsm representation
            To be overloaded by subclasses.
        """
        raise NotImplementedError()

    def get_expected_dataflow(self):
        """Returns the expected data exchange sequence,
        uUsed for xdsm representation
        to be overloaded by subclasses

        :returns: array of tuples (disc_from, disc_to, array of variable names)
        """
        raise NotImplementedError()

    @classmethod
    def get_default_sub_options_values(cls, **options):  # pylint: disable=W0613
        """
        When some options of the formulation depend on higher level
        options, a sub option defaults may be specified here, mainly for
        use in the API

        :param options: options dict required to deduce the sub options grammar
        :returns: None, or the sub options defaults
        """

    @classmethod
    def get_sub_options_grammar(cls, **options):  # pylint: disable=W0613
        """
        When some options of the formulation depend on higher level
        options, a sub option schema may be specified here, mainly for
        use in the API

        :param options: options dict required to deduce the sub options grammar
        :returns: None, or the sub options grammar
        """



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Matthias De Lozzo, Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Min-max data scaler
===================

The :class:`.MinMaxScaler` class implements the MinMax scaling method applying
to some parameter :math:`z`:

.. math::

    \\bar{z} := \\text{offset} + \\text{coefficient}\\times z
    = \\frac{z-\\text{min}(z)}{(\\text{max}(z)-\\text{min}(z))},

where :math:`\\text{offset}=-\\text{min}(z)/(\\text{max}(z)-\\text{min}(z))`
and :math:`\\text{coefficient}=1/(\\text{max}(z)-\\text{min}(z))`.

In the MinMax scaling method, the scaling operation linearly transforms the
original variable :math:`z` such that the minimum of the original data
corresponds to 0 and the maximum to 1.
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library

from gemseo.mlearning.transform.scaler.scaler import Scaler

standard_library.install_aliases()


class MinMaxScaler(Scaler):
    """ Min-max scaler. """

    def __init__(self, name="MinMaxScaler", offset=0.0, coefficient=1.0):
        """Constructor.

        :param str name: name of the scaler. Default: 'MinMaxScaler'.
        :param float offset: offset of the linear transformation.
            Default: 0.
        :param float coefficient: coefficient of the linear transformation.
            Default: 1.
        """
        super(MinMaxScaler, self).__init__(name, offset, coefficient)

    def fit(self, data):
        """Fit offset and coefficient terms from a data array. The min and the
        max are computed along the first axis of the data.

        :param array data: data to be fitted.
        """
        super(MinMaxScaler, self).fit(data)
        l_b = data.min(0)
        u_b = data.max(0)
        self.offset = -l_b / (u_b - l_b)
        self.coefficient = 1 / (u_b - l_b)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Data transformer pipeline
=========================

The :class:`.Pipeline` class chains a sequence of tranformers, and provides
global fit(), transform(), fit_transform() and inverse_transform() methods.
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library
from numpy import eye, matmul

from gemseo.mlearning.transform.transformer import Transformer

standard_library.install_aliases()


class Pipeline(Transformer):
    """ Transformer pipeline. """

    def __init__(self, name="Pipeline", transformers=None):
        """Constructor.

        :param str name: transformer pipeline name. Default: 'Pipeline'.
        :param list(Transformer) transformers: Sequence of transformers to be
            chained. The transformers are chained in the order of appearance in
            the list, i.e. the first transformer is applied first. If
            transformers is an empty list or None, then the pipeline
            transformer behaves like an identity transformer.
            Default: None.
        """
        super(Pipeline, self).__init__(name)
        self.transformers = transformers or []

    def duplicate(self):
        """ Duplicate the constructor. """
        transformers = [trans.duplicate() for trans in self.transformers]
        return Pipeline(self.name, transformers)

    def fit(self, data):
        """Fit transformer pipeline to data. All the transformers are fitted,
        transforming the data along the way.

        :param ndarray data: data to be fitted.
        """
        for transformer in self.transformers:
            data = transformer.fit_transform(data)

    def transform(self, data):
        """Transform data. The data is transformed sequentially, where the
        output of one transformer is the input of the next.

        :param ndarray data: data to be transformed.
        :return: transformed data.
        :rtype: ndarray
        """
        for transformer in self.transformers:
            data = transformer.transform(data)
        return data

    def inverse_transform(self, data):
        """Perform an inverse transform on the data. The data is inverse
        transformed sequentially, starting with the last tranformer in the
        list.

        :param ndarray data: data  to be inverse transformed.
        :return: inverse transformed data.
        :rtype: ndarray
        """
        for transformer in self.transformers[::-1]:
            data = transformer.inverse_transform(data)
        return data

    def compute_jacobian(self, data):
        """Compute Jacobian of the pipeline transform.

        :param ndarray data: data where the Jacobian is to be computed.
        :return: Jacobian matrix.
        :rtype: ndarray
        """
        jacobian = eye(data.shape[-1])
        for transformer in self.transformers:
            jacobian = matmul(transformer.compute_jacobian(data), jacobian)
            data = transformer.transform(data)
        return jacobian

    def compute_jacobian_inverse(self, data):
        """Compute Jacobian of the pipeline inverse_transform.

        :param ndarray data: data where the Jacobian is to be computed.
        :return: Jacobian matrix.
        :rtype: ndarray
        """
        jacobian = eye(data.shape[-1])
        for transformer in self.transformers[::-1]:
            jacobian = matmul(transformer.compute_jacobian_inverse(data), jacobian)
            data = transformer.inverse_transform(data)
        return jacobian



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Parameter space including both deterministic and uncertain parameters
=====================================================================

Overview
--------

The :class:`.ParameterSpace` class describes a set of parameters of
interest which can be either deterministic or uncertain. This class
inherits from :class:`.DesignSpace`.

Capabilities
------------

The :meth:`.DesignSpace.add_variable` aims to add deterministic
variables from:

- a variable name,
- a variable size (default: 1),
- a variable type (default: float),
- a lower bound (default: - infinity),
- an upper bound (default: + infinity),
- a current value (default: None).

The :meth:`.add_random_variable` aims to add uncertain
variables (a.k.a. random variables) from:

- a variable name,
- a distribution name
  (see :meth:`~gemseo.uncertainty.api.get_available_distributions`),
- a variable size,
- distribution parameters (:code:`parameters` set as
  a tuple of positional arguments for :class:`.OTDistribution`
  or a dictionary of keyword arguments for :class:`.SPDistribution`,
  or keyword arguments for standard probability distribution such
  as :class:`.OTNormalDistribution` and :class:`.OTNormalDistribution`).

The :class:`.ParameterSpace` also provides the following methods:

- :meth:`.get_cdf`: evaluate the cumulative density function
  for the different variables and their different
- :meth:`.get_composed_distribution`: returns the probability distribution
  of an uncertain variable,
- :meth:`.get_marginal_distributions`: returns the marginal probability
  distributions of an uncertain variable,
- :meth:`.get_range` returns the numerical range
  of the different uncertain parameters,
- :meth:`.ParameterSpace.get_sample`: returns several sample
  of the uncertain variables,
- :meth:`.get_support`: returns the mathematical support
  of the different uncertain variables,
- :meth:`.is_uncertain`: checks if a parameter is uncertain,
- :meth:`.is_deterministic`: checks if a parameter is deterministic.
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library
from numpy import array, ndarray
from sympy import simplify

from gemseo.algos.design_space import DesignSpace
from gemseo.uncertainty.distributions.distribution import ComposedDistribution
from gemseo.uncertainty.distributions.factory import DistributionFactory
from gemseo.utils.data_conversion import DataConversion

standard_library.install_aliases()

from gemseo import LOGGER


class ParameterSpace(DesignSpace):
    """ Parameter space. """

    INITIAL_DISTRIBUTION = "Initial distribution"
    TRANSFORMATION = "Transformation"
    SUPPORT = "Support"
    MEAN = "Mean"
    STANDARD_DEVIATION = "Standard deviation"
    RANGE = "Range"
    BLANK = ""
    PARAMETER_SPACE = "Parameter space"

    def __init__(
        self,
        print_decimals=2,
        shorten=True,
        copula=ComposedDistribution.INDEPENDENT_COPULA,
    ):
        """Constructor

        :param int print_decimals: number of decimals to print. Default: 2.
        :param bool shorten: if True, simplify the expressions
            of variable transformations. Default: True.
        :param str copula: copula name.
            Default: ComposedDistribution.INDEPENDENT_COPULA.
        """
        LOGGER.info("Create a new parameter space. ")
        super(ParameterSpace, self).__init__()
        self.uncertain_variables = []
        self.marginals = {}
        self.distribution = None
        self._ndecimals = print_decimals
        self._shorten = shorten
        if copula not in ComposedDistribution.AVAILABLE_COPULA:
            raise ValueError("%s is not an available copula." % (copula))
        self.copula = copula

    def is_uncertain(self, variable):
        """Check if a variable is uncertain.

        :param str variable: variable name.
        """
        return variable in self.uncertain_variables

    def is_deterministic(self, variable):
        """Check if a variable is deterministic.

        :param str variable: variable name.
        """
        determistic = set(self.variables_names) - set(self.uncertain_variables)
        return variable in determistic

    def __update_parameter_space(self, variable):
        """Update parameter space.

        :param variable: variable name.
        :type variable: str
        """
        if variable not in self.variables_names:
            l_b = self.marginals[variable].math_lower_bound
            u_b = self.marginals[variable].math_upper_bound
            value = self.marginals[variable].mean
            size = self.marginals[variable].dimension
            self.add_variable(variable, size, "float", l_b, u_b, value)
        else:
            l_b = self.marginals[variable].math_lower_bound
            u_b = self.marginals[variable].math_upper_bound
            value = self.marginals[variable].mean
            self.set_lower_bound(variable, l_b)
            self.set_upper_bound(variable, u_b)
            self.set_current_variable(variable, value)

    def add_random_variable(self, name, distribution, size=1, **parameters):
        """Add a random variable from a distribution

        :param str name: name of the random variable.
        :param str distribution: distribution name.
        :param int size: variable size.
        :param parameters: parameters of the distribution.
        """
        factory = DistributionFactory()
        distribution = factory.create(
            distribution, variable=name, dimension=size, **parameters
        )
        variable = distribution.variable_name
        LOGGER.info("Add the random variable: %s", variable)
        self.marginals[variable] = distribution
        self.uncertain_variables.append(variable)
        self._build_composed_distribution()
        self.__update_parameter_space(variable)

    def _build_composed_distribution(self):
        """ Build the composed distribution from the marginal ones. """
        tmp_marginal = self.marginals[self.uncertain_variables[0]]
        marginals = [self.marginals[name] for name in self.uncertain_variables]
        self.distribution = tmp_marginal.COMPOSED_DISTRIBUTION(marginals, self.copula)

    def get_composed_distribution(self, variable):
        """Get the composed distribution of a random variable.

        :param str variable: variable name.
        """
        return self.marginals[variable].distribution

    def get_marginal_distributions(self, variable):
        """Get the marginal distributions of a random variable.

        :param str variable: variable name.
        """
        return self.marginals[variable].marginals

    def get_range(self, variable):
        """Get the numerical range of a random variable.

        :param str variable: variable name.
        """
        return self.marginals[variable].range

    def get_support(self, variable):
        """Get the mathematical support of a random variable.

        :param str variable: variable name.
        """
        return self.marginals[variable].support

    def remove_variable(self, name):
        """Remove a variable from the probability space.

        :param str name: variable name.
        """
        if name in self.uncertain_variables:
            del self.marginals[name]
            self.uncertain_variables.remove(name)
            self._build_composed_distribution()
        super(ParameterSpace, self).remove_variable(name)

    def set_dependence(self, variables, copula, **options):
        """Set dependence relation between random variables.

        :param list(str) variables: list of variables names.
        :param str copula: copula name.
        :param options: copula options.
        """
        raise NotImplementedError

    def get_sample(self, n_samples=1, as_dict=False):
        """Get sample.

        :param int n_samples: number of samples.
        :param bool as_dict: return a dictionary.
        :return: samples
        :rtype: list(array) or list(dict)
        """
        sample = self.distribution.get_sample(n_samples)
        if as_dict:
            sample = [
                DataConversion.array_to_dict(
                    data_array, self.uncertain_variables, self.variables_sizes
                )
                for data_array in sample
            ]
        return sample

    def get_cdf(self, value, inverse=False):
        """Get the inverse Cumulative Density Function
         values of the different marginals.

        :param dict(array) value: values
        :return: (inverse) CDF values
        :rtype: dict(array)
        """
        if inverse:
            self.__check_dict_of_array(value)
        values = {}
        for name in self.uncertain_variables:
            val = value[name]
            distribution = self.marginals[name]
            if inverse:
                current_v = distribution.inverse_cdf(val)
            else:
                current_v = distribution.cdf(val)
            values[name] = array(current_v)

        return values

    def __check_dict_of_array(self, obj):
        """Check if the object is a dictionary of array.

        :param obj: object to test
        """
        error_msg = "obj must be a dictionary whose keys are the variables "
        error_msg += "names and values are arrays whose dimensions are the "
        error_msg += "variables ones and components are in [0, 1]."
        if not isinstance(obj, dict):
            raise TypeError(error_msg)
        for variable, value in obj.items():
            if variable not in self.uncertain_variables:
                LOGGER.debug(
                    "%s is not defined in the probability space."
                    " Available variables are [%s]."
                    " Use uniform distribution for %s.",
                    variable,
                    ", ".join(self.uncertain_variables),
                    variable,
                )
            else:
                if not isinstance(value, ndarray):
                    raise TypeError(error_msg)
                if len(value.flatten()) != self.variables_sizes[variable]:
                    raise ValueError(error_msg)
                if any(value.flatten() > 1.0) or any(value.flatten() < 0.0):
                    raise ValueError(error_msg)

    def __str__(self, *args, **kwargs):
        """String representation.

        :return: description
        :rtype: str
        """
        table = self.get_pretty_table()
        distribution = []
        transformation = []
        support = []
        mean = []
        std = []
        rnge = []
        for variable in self.variables_names:
            if variable in self.uncertain_variables:
                dist = self.marginals[variable]
                tmp_mean = dist.mean
                tmp_std = dist.standard_deviation
                tmp_range = dist.range
                tmp_support = dist.support
                for dim in range(dist.dimension):
                    distribution.append(str(dist))
                    transformation.append(dist.transformation)
                    if self._shorten:
                        transformation[-1] = str(simplify(transformation[-1]))
                    mean.append(tmp_mean[dim])
                    mean[-1] = round(mean[-1], self._ndecimals)
                    std.append(tmp_std[dim])
                    std[-1] = round(std[-1], self._ndecimals)
                    rnge.append(tmp_range[dim])
                    support.append(tmp_support[dim])
            else:
                for dim in range(self.variables_sizes[variable]):
                    distribution.append(self.BLANK)
                    transformation.append(self.BLANK)
                    mean.append(self.BLANK)
                    std.append(self.BLANK)
                    support.append(self.BLANK)
                    rnge.append(self.BLANK)

        table.add_column(self.INITIAL_DISTRIBUTION, distribution)
        table.add_column(self.TRANSFORMATION, transformation)
        table.add_column(self.SUPPORT, support)
        table.add_column(self.MEAN, mean)
        table.add_column(self.STANDARD_DEVIATION, std)
        table.add_column(self.RANGE, rnge)
        table.title = self.PARAMETER_SPACE
        desc = str(table)
        return desc

    def unnormalize_vect(self, x_vect, minus_lb=True, no_check=False, use_dist=True):
        """Inverse transformation from a unit design vector.
        Unnormalizes a normalized vector of the design space.

        :param array x_vect: design variables.
        :param bool minus_lb: if True, remove lower bounds at normalization.
        :param bool no_check: if True, don't check that values are in [0,1].
        :param bool use_dist: if True, rescale wrt the stats law.
        :return: normalized vector
        :rtype: array
        """
        if not use_dist:
            return super(ParameterSpace, self).unnormalize_vect(x_vect)

        data_names = self.variables_names
        data_sizes = self.variables_sizes
        dict_sample = DataConversion.array_to_dict(x_vect, data_names, data_sizes)
        x_u_geom = super(ParameterSpace, self).unnormalize_vect(x_vect)
        x_u = self.get_cdf(dict_sample, inverse=True)
        x_u_geom = DataConversion.array_to_dict(x_u_geom, data_names, data_sizes)
        missing_names = list(set(data_names) - set(x_u.keys()))
        for name in missing_names:
            x_u[name] = x_u_geom[name]
        x_u = DataConversion.dict_to_array(x_u, data_names)
        return x_u

    def normalize_vect(self, x_vect, minus_lb=True, use_dist=False):
        """Normalizes a vector of the design space.
        Unbounded variables are not normalized.

        :param array x_vect: design variables.
        :param bool minus_lb: if True, remove lower bounds at normalization.
        :param bool no_check: if True, don't check that values are in [0,1].
        :param bool use_dist: if True, rescale wrt the stats law.
        :return: normalized vector
        :rtype: array
        """
        if use_dist:
            return super(ParameterSpace, self).normalize_vect(x_vect)

        data_names = self.variables_names
        data_sizes = self.variables_sizes
        dict_sample = DataConversion.array_to_dict(x_vect, data_names, data_sizes)
        x_u_geom = super(ParameterSpace, self).normalize_vect(x_vect)
        x_u = self.get_cdf(dict_sample, inverse=False)
        x_u_geom = DataConversion.array_to_dict(x_u_geom, data_names, data_sizes)
        missing_names = list(set(data_names) - set(x_u.keys()))
        for name in missing_names:
            x_u[name] = x_u_geom[name]
        x_u = DataConversion.dict_to_array(x_u, data_names)
        return x_u



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or
#                      initial documentation
#        :author:  Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Make a discipline from an executable
************************************
"""
from __future__ import absolute_import, division, unicode_literals

import re
import subprocess
from ast import literal_eval
from copy import deepcopy
from multiprocessing import Lock, Manager
from os import listdir, mkdir
from os.path import join
from uuid import uuid1

from future import standard_library
from numpy import array

from gemseo import SOFTWARE_NAME
from gemseo.api import configure_logger
from gemseo.core.data_processor import FloatDataProcessor
from gemseo.core.discipline import MDODiscipline
from gemseo.core.json_grammar import JSONGrammar
from gemseo.utils.py23_compat import OrderedDict  # automatically dict from py36
from gemseo.utils.py23_compat import xrange

standard_library.install_aliases()

configure_logger(SOFTWARE_NAME)

from gemseo import LOGGER

INPUT_TAG = "GEMSEO_INPUT"
OUTPUT_TAG = "GEMSEO_OUTPUT"
INPUT_RE = INPUT_TAG + r"\{(.*)\}"
OUTPUT_RE = OUTPUT_TAG + r"\{(.*)\}"

INPUT_GRAMMAR = JSONGrammar.INPUT_GRAMMAR
OUTPUT_GRAMMAR = JSONGrammar.OUTPUT_GRAMMAR

NUMERICS = [str(j) for j in xrange(10)]


class DiscFromExe(MDODiscipline):
    """Generic wrapper for executables.

    The DiscFromExe is a generic wrapper for executables. It generates a
    MDODiscipline from an executable and in inputs/output files wrappers.  The
    input and output files are described by templates.  The templates can be
    generated by executing the module
    :mod:`~gemseo.wrappers.template_grammar_editor` to open a GUI.


    It requires the creation of templates for input and output file,
    for instance, from the following input JSON file:

    .. code::

        {
        "a": 1.01515112125,
        "b": 2.00151511213,
        "c": 3.00151511213
        }

    A template that declares the inputs must be generated under this format,
    where "a" is the name of the input, and "1.0" is the default input.
    GEMSEO_INPUT declares an input, GEMSEO_OUTPUT declares an output, similarly.

    .. code::

        {
        "a": GEMSEO_INPUT{a::1.0},
        "b": GEMSEO_INPUT{b::2.0},
        "c": GEMSEO_INPUT{c::3.0}
        }


    Current limitations :

        Only one input and one output file, otherwise, inherit from this class
        and modify the parsers.
        Only limited input writing and output parser strategies
        are implemented. To change that, you can pass custom parsing and
        writing methods to the constructor.

        The only limitation in the current file format is that
        it must be a plain text file and not a binary file.
        In this case, the way of interfacing it is
        to provide a specific parser to the DiscFromExe,
        with the write_input_file_method
        and parse_outfile_method arguments of the constructor.

    """

    NUMBERED = "numbered"
    UUID = "uuid"
    KEY_VALUE_PARSER = "KEY_VALUE_PARSER"
    TEMPLATE_PARSER = "TEMPLATE_PARSER"

    def __init__(
        self,
        input_template,
        output_template,
        output_folder_basepath,
        executable_command,
        input_filename,
        output_filename,
        folders_iter=NUMBERED,
        name=None,
        parse_outfile_method=TEMPLATE_PARSER,
        write_input_file_method=None,
        parse_out_separator="=",
    ):
        """Constructor.

        Create the discipline from the inputs, outputs wrapper and
        the executable command.

        :param str input_template: path to the input file template.
            The input locations in the file are marked
            by GEMSEO_INPUT{input_name::1.0},
            where "input_name" is the input name, and 1.0 is here
            the default input
        :param str output_template: path to the output file template.
            The input locations in the file are marked
            by GEMSEO_OUTPUT{output_name::1.0},
            where "output_name" is the input name
        :param str output_folder_basepath: path to the output folder,
            in which the executions will be performed
        :param str executable_command: command to run the executable.
            Will be called through a system call.
            Example: "python myscript.py -i input.txt -o output.txt
        :param str input_filename: name of the input file.
            This will determine the name
            of the input file generated in the output folder.
            Example "input.txt"
        :param str output_filename: name of the output file.
            This will determine the name
            of the output file generated in the output folder.
            Example "output.txt"
        :param str folders_iter: type of unique identifiers for the output
            folders. If NUMBERED the generated output folders
            will be "output_folder_basepath"+str(i+1),
            where i is the maximum value of the already existing
            "output_folder_basepath"+str(i) folders.
            Otherwise, a unique number based on the UUID function is
            generated. This last option shall be used if multiple MDO
            processes are runned in the same work directory.
        :param str parse_outfile_method: optional method that can be provided
            by the user to parse the output file. To see the signature of
            the method, see the parse_outfile method of this file.
        :param str parse_out_separator: if the KEY_VALUE_PARSER is used as
            output parser, specify the separator key (default : "=").
        :param str write_input_file_method: method to write the input file,
            if None, use this modules' write_input_file. To see the signature
            of the method, see the write_input_file method of this file.
        """

        super(DiscFromExe, self).__init__(name=name)
        self.input_template = input_template
        self.output_template = output_template
        self.input_filename = input_filename
        self.output_filename = output_filename
        self.executable_command = executable_command

        use_template_parse = parse_outfile_method == self.TEMPLATE_PARSER
        if parse_outfile_method is None or use_template_parse:
            self.parse_outfile = parse_outfile
        elif parse_outfile_method == self.KEY_VALUE_PARSER:
            self.parse_outfile = lambda a, b: parse_key_value_file(
                a, b, parse_out_separator
            )
        else:
            self.parse_outfile = parse_outfile_method
        if not hasattr(self.parse_outfile, "__call__"):
            raise TypeError("The parse_outfile_method must be callable")

        self.write_input_file = write_input_file_method or write_input_file
        if not hasattr(self.write_input_file, "__call__"):
            raise TypeError("The write_input_file_method must be callable")
        self.lock = Lock()
        self.folders_iter = folders_iter
        self.output_folder_basepath = output_folder_basepath

        self._out_pos = None
        self._in_dict = None
        self._out_dict = None
        self._in_lines = None
        self._out_lines = None

        n_dirs = self._get_max_outdir()
        self.counter = Manager().Value("i", n_dirs)

        self.data_processor = FloatDataProcessor()

        self._parse_templates()

    def _parse_templates(self):
        """Parse the templates.

        Parse the templates and:
            Initialize the grammars
            Initialize the attributes : self._in_lines, self._out_lines,
            self._out_pos, self._out_pos
            self.default_inputs

        """
        with open(self.input_template, "r") as infile:
            self._in_lines = infile.readlines()
        with open(self.output_template, "r") as outfile:
            self._out_lines = outfile.readlines()

        self._in_dict, self._in_pos = parse_template(self._in_lines, INPUT_GRAMMAR)
        self.input_grammar.initialize_from_data_names(self._in_dict.keys())

        out_dict, self._out_pos = parse_template(self._out_lines, OUTPUT_GRAMMAR)

        self.output_grammar.initialize_from_data_names(out_dict.keys())

        msg = "Initialize discipline from template. \
                Input grammar: {}".format(
            self._in_dict.keys()
        )
        LOGGER.debug(msg)
        msg = "Initialize discipline from template. \
                Output grammar: {}".format(
            out_dict.keys()
        )
        LOGGER.debug(msg)
        self.default_inputs = {
            k: array([literal_eval(v)]) for k, v in self._in_dict.items()
        }

    def _run(self):
        """Run the wrapper."""
        uuid = self.generate_uid()

        out_dir = join(self.output_folder_basepath, uuid)

        mkdir(out_dir)
        input_file_path = join(out_dir, self.input_filename)

        self.write_input_file(
            input_file_path, self.local_data, self._in_pos, self._in_lines
        )

        err = subprocess.call(
            self.executable_command, shell=True, stderr=subprocess.STDOUT, cwd=out_dir
        )
        if err != 0:
            raise RuntimeError("Execution failed and returned error code : " + str(err))
        outfile = join(out_dir, self.output_filename)
        with open(outfile, "r") as outfile:
            out_lines = outfile.readlines()

        if len(out_lines) != len(self._out_lines):
            raise ValueError(
                "The number of lines of the output file changed."
                "This is not supported yet"
            )

        out_vals = self.parse_outfile(self._out_pos, out_lines)
        self.local_data.update(out_vals)

    def generate_uid(self):
        """Generate an UUID.

        Generate a unique identifier for the current execution If the
        folders_iter strategy is NUMBERED, the successive iterations are named
        by an integer 1, 2, 3 etc. This is multiprocess safe.  Otherwise, a
        unique number based on the UUID function is generated.  This last
        option shall be used if multiple MDO processes are runned in the same
        workdir.

        :returns: a unique string identifier
        :rtype: str
        """
        if self.folders_iter == self.NUMBERED:
            with self.lock:
                self.counter.value += 1
                return str(self.counter.value)
        return str(uuid1()).split("-")[-1]

    def _list_out_dirs(self):
        """List the directories in the output folder path."""
        return listdir(self.output_folder_basepath)

    def _get_max_outdir(self):
        """Get the maximum current index of output folders."""
        outs = list(self._list_out_dirs())
        if not outs:
            return 0
        return max([literal_eval(n) for n in outs])


def parse_template(template_lines, grammar_type=INPUT_GRAMMAR):
    """Parse the input or output template.

    :param template_lines: list of lines of the file template
        (result of file.readlines())
    :param grammar_type: INPUT_GRAMMAR or OUTPUT_GRAMMAR
    :returns: data_dict, pos_dict, where data_dict is the {name:value} dict,
        where name is the data name and value is the parsed input or output
        value in the template pos_dict in the format dictionary
        containing the information from the template
        format {data_name:(start,end,line_number)}, where
        name is the name of the input
        data, start is the index of the starting point
        in the input file template.
        This index is a line index (character number on the line)
        end is the index of the end character in the template
        line_number is the index of the line in the file
    """
    if grammar_type == INPUT_GRAMMAR:
        pattern_re = INPUT_RE
    elif grammar_type == OUTPUT_GRAMMAR:
        pattern_re = OUTPUT_RE
    else:
        raise ValueError("Unknown grammar type " + str(grammar_type))
    regex = re.compile(pattern_re)  # , re.MULTILINE
    data_dict = OrderedDict()
    pos_dict = OrderedDict()

    for lineid, line in enumerate(template_lines):
        for match in regex.finditer(line):
            data = match.groups()[0]
            spl = data.split("::")
            name = spl[0]
            val = spl[1]
            data_dict[name] = val
            # When input mode: erase the template value
            if grammar_type == INPUT_GRAMMAR:
                start, end = match.start(), match.end()
            else:
                # In output mode : catch all
                # the output lenght and not more
                start = match.start()
                end = start + len(val)

            pos_dict[name] = (start, end, lineid)

    return data_dict, pos_dict


def write_input_file(
    input_file_path, data, input_positions, input_lines, float_format="{:1.18g}"
):
    """Write the input file from the input data.

    :param input_file_path: absolute path to the file to be written
    :param data: data dictionary, ie the local data of the discipline
    :param input_positions: dictionary containing the information
        from the template format {data_name:(start,end,line_number)}, where
        name is the name of the input
        data, start is the index of the starting point
        in the input file template.
        This index is a line index (character number on the line)
        end is the index of the end character in the template
        line_number is the index of the line in the file
    :param input_lines: list of lines of the input file template
        (result of file.readlines())
    :param float_format: formating of the input data in the file
    """
    f_text = deepcopy(input_lines)
    for name, pos in input_positions.items():
        start, end, lineid = pos
        data_str = float_format.format(data[name])
        cline = f_text[lineid]
        f_text[lineid] = cline[:start] + data_str + cline[end:]

    with open(input_file_path, "w") as infile_o:
        infile_o.writelines(f_text)


def parse_key_value_file(_, out_lines, separator="="):
    """Parse the output file from the expected text positions.

    :param out_lines: list of lines of the output file template
        (result of file.readlines())
    :param separator: separating characters of the key=value format
    :returns: the values dictionary in dict of numpy array formats
    """
    data = {}
    for line in out_lines:
        if separator in line:
            spl = line.strip().split(separator)
            if len(spl) != 2:
                raise ValueError("unbalanced = in line " + str(line))
            key = spl[0].strip()
            try:
                data[key] = float(literal_eval(spl[1].strip()))
            except Exception:
                raise ValueError("Failed to parse value as float " + str(spl[1]))

    return data


def parse_outfile(output_positions, out_lines):
    """Parse the output file from the expected text positions.

    :param output_positions: dictionary containing the information
        from the template format {data_name:(start,end,dictionary)}, where
        name is the name of the output
        data, start is the index of the starting point
        in the input file template.
        This index is a line index (character number on the line)
        end is the index of the end character in the template
        line_number is the index of the line in the file
    :param out_lines: list of lines of the output file template
        (result of file.readlines())
    :returns: the values dictionary in dict of numpy array formats
    """
    values = {}
    for name, pos in output_positions.items():
        start, _, lineid = pos

        found_dot = False
        found_e = False
        # In case generated files has less lines
        if lineid > len(out_lines) - 1:
            break
        out_text = out_lines[lineid]
        i = start
        maxi = len(out_text)
        while True:
            # The problem is that the output file used for the template may be
            # using a output that is longer or shorter than the one generated
            # at runtime. We must find the proper end of the expression...
            i += 1
            char = out_text[i]
            if char == ".":
                # We found the . in float notation
                if found_dot or found_e:
                    break
                found_dot = True
                continue
            # We found the e in exp notation
            if char in ("E", "e"):
                if found_e:
                    break
                found_e = True
                continue
            # Check that we have nout reached EOL or space or whatever
            if char not in NUMERICS:
                break
            if i == maxi - 1:
                print("IBRAEAU")
                break
        outv = out_text[pos[0] : i]

        LOGGER.info('Parsed "' + name + '" got output ' + outv)
        values[name] = array([float(outv)])
    return values



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Various termination criteria for drivers
****************************************
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from future import standard_library

standard_library.install_aliases()


class TerminationCriterion(Exception):
    """Stop driver for some reason"""


class FunctionIsNan(TerminationCriterion):
    """Stops driver when a function has NaN value or NaN Jacobian"""


class DesvarIsNan(TerminationCriterion):
    """Stops driver when the design variables are nan"""


class MaxIterReachedException(TerminationCriterion):
    """
    Exception raised when the maximum number of iterations is reached
    by the driver
    """



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
K-means clustering algorithm
============================

The k-means algorithm groups the data into clusters, where the number of
clusters :math:`k` is fixed. This is done by initializing :math:`k` centroids
in the design space. The points are grouped into clusters according to their
nearest centroid.

When fitting the algorithm, each centroid is successively moved to the mean of
its corresponding cluster, and the cluster value of each point is then reset to
the cluster value of the closest centroid. This process is repeated until
convergence.

Cluster values of new points may be predicted by returning the value of the
closest centroid. Denoting
:math:`(c_1, \\cdots, c_k) \\in \\mathbb{R}^{n \\times k}` the centroids, and
assuming no overlap between the centroids, we may compute the prediction

.. math::

        \\operatorname{cluster}(x) =
            \\underset{i=1,\\cdots,k}{\\operatorname{argmin}} \\|x-c_i\\|.

A probability measure may also be provided, using the distances from the point
to each of the centroids:

.. math::

    \\mathbb{P}(x \\in C_i) = \\begin{cases}
        1 & \\operatorname{if} x = c_i\\\\
        0 & \\operatorname{if} x = c_j,\\ j \\neq i\\\\
        \\frac{\\frac{1}{\\|x-c_i\\|}}{\\sum_{j=1}^k \\frac{1}{\\|x-c_j\\|}}
            & \\operatorname{if} x \\neq c_j\\, \\forall j=1,\\cdots,k
    \\end{cases},

where :math:`C_i = \\{x\\, | \\, \\operatorname{cluster}(x) = i \\}`. Here,
:math:`\\mathbb{P}(x \\in C_i)` represents the probability of cluster :math:`i`
given the point :math:`x`.


This concept is implemented through
the :class:`.KMeans` class which inherits from
the :class:`.MLClusteringAlgo` class.

Dependence
----------
This clustering algorithm relies on the KMeans class
of the `scikit-learn library <https://scikit-learn.org/stable/modules/
generated/sklearn.cluster.KMeans.html>`_.
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library
from numpy import finfo
from numpy.linalg import norm
from sklearn.cluster import KMeans as SKLKmeans

from gemseo.mlearning.cluster.cluster import MLClusteringAlgo

standard_library.install_aliases()


from gemseo import LOGGER


class KMeans(MLClusteringAlgo):
    """ KMeans clustering algorithm. """

    ABBR = "KMeans"

    EPS = finfo(float).eps

    def __init__(
        self,
        data,
        transformer=None,
        var_names=None,
        n_clusters=5,
        random_state=0,
        **parameters
    ):
        """Constructor.

        :param data: learning dataset.
        :type data: Dataset
        :param transformer: transformation strategy for data groups.
            If None, do not transform data. Default: None.
        :type transformer: dict(str)
        :param var_names: names of the variables to consider.
        :type var_names: list(str)
        :param n_clusters: number of clusters. Default: 5.
        :type n_clusters: int
        :param random_state: If None, use a random generation
            of the initial centroids.
            Use an int to make the randomness deterministic. Default: 0.
        :type random_state: int
        :param parameters: Scikit-learn algorithm parameters.
        """
        super(KMeans, self).__init__(
            data,
            transformer=transformer,
            var_names=var_names,
            n_clusters=n_clusters,
            random_state=random_state,
            **parameters
        )
        self.algo = SKLKmeans(n_clusters, random_state=random_state, **parameters)

    def _fit(self, data):
        """Fit the clustering model to the data and store labels.

        :param ndarray data: training data (2D).
        """
        self.labels = self.algo.fit_predict(data)

    def _predict(self, data):
        """Predict cluster of data.

        :param ndarray data: data (2D).
        :return: clusters of data (1D).
        :rtype: ndarray(int).
        """
        return self.algo.predict(data)

    def _predict_proba_soft(self, data):
        """Predict probability of belonging to each cluster.

        :param ndarray data: data (2D).
        :return: probabilities for each cluster for each sample (2D). The sum
            of each row is one.
        :rtype: ndarray
        """
        centers = self.algo.cluster_centers_
        distances = norm(data[:, None] - centers, axis=2)
        inverse_distances = 1 / (distances + self.EPS)
        probas = inverse_distances / inverse_distances.sum(axis=1)[:, None]
        return probas
